from abc import ABCMeta
import argparse
from clipzyme.utils.registry import get_object

INITED_OBJ = []


class classproperty(object):
    """
    Method decorator behaves as @classmethod + @property
    """

    def __init__(self, fget):
        self.fget = fget

    def __get__(self, owner_self, owner_cls):
        return self.fget(owner_cls)


class Nox(object):
    __metaclass__ = ABCMeta

    def __init__(self, **kwargs) -> None:
        super(Nox, self).__init__()

    @staticmethod
    def add_args(parser) -> None:
        """Add class specific args

        Args:
            parser (argparse.ArgumentParser): argument parser
        """
        pass

    @staticmethod
    def set_args(args) -> None:
        """Set values for class specific args

        Args:
            args (argparse.Namespace): arguments
        """
        pass


def set_nox_type(object_name):
    """
    Build argparse action class for registry items
    Used to add and set object-level args

    Args:
        object_name (str): kind of nox class uses (e.g., dataset, model, lightning)

    Returns:
        argparse.Action: action for specific nox class
    """

    class NoxAction(argparse.Action):
        def __init__(
            self,
            option_strings,
            dest,
            nargs=None,
            const=None,
            default=None,
            type=None,
            choices=None,
            required=False,
            help=None,
            metavar=None,
        ):
            super().__init__(
                option_strings=option_strings,
                dest=dest,
                nargs=nargs,
                const=const,
                default=default,
                type=type,
                choices=choices,
                required=required,
                help=help,
                metavar=metavar,
            )
            self.is_nox_action = True
            self.object_name = object_name

        def __call__(self, parser, namespace, values, option_string=None) -> None:
            setattr(namespace, self.dest, values)

        def add_args(self, parser, values) -> None:
            """
            Add object-level args when an add_argument is called

            Args:
                parser (argparse.parser): nox parser object
                values (Union[list, str]): argument values inputted
            """
            if isinstance(values, list):
                for v in values:
                    obj_val_str = f"{v}_{object_name}"
                    # if object has already been called, conflict arises with add parse called multiple times
                    if obj_val_str not in INITED_OBJ:
                        get_object(v, object_name).add_args(parser)
                        INITED_OBJ.append(obj_val_str)

            elif isinstance(values, str):
                obj_val_str = f"{values}_{object_name}"
                # if object has already been called, conflict arises with add parse called multiple times
                if obj_val_str not in INITED_OBJ:
                    get_object(values, object_name).add_args(parser)
                    INITED_OBJ.append(obj_val_str)

        def set_args(self, args, val) -> None:
            """
            Call object-level set_args method

            Args:
                args (argparse.namespace): global args
                val (Union[list,str]): value for argument
            """
            if isinstance(val, list):
                for v in val:
                    get_object(v, object_name).set_args(args)
            elif isinstance(val, str):
                get_object(val, object_name).set_args(args)

    return NoxAction



from clipzyme.utils.registry import register_object
import pytorch_lightning as pl
import os
from clipzyme.utils.classes import Nox


@register_object("tensorboard", "logger")
class PLTensorBoardLogger(pl.loggers.TensorBoardLogger, Nox):
    def __init__(self, args) -> None:
        super().__init__(args.logger_dir)

    def setup(self, **kwargs):
        pass

    def log_image(self, image, name):
        pass

    @staticmethod
    def add_args(parser) -> None:
        parser.add_argument(
            "--logger_dir",
            type=str,
            default=".",
            help="directory to save tensorboard logs",
        )



from __future__ import absolute_import, division, print_function, unicode_literals
from typing import Union, Tuple, Any, List
import numpy as np
import re
import warnings
from functools import partial

from rdkit import Chem, DataStructs, rdBase
from rdkit.Chem import AllChem
from rdkit.Chem.Scaffolds import MurckoScaffold

from rxn.chemutils.conversion import canonicalize_smiles, smiles_to_mol
from rxn.chemutils.exceptions import InvalidSmiles
from rxn.chemutils.utils import remove_atom_mapping
from rxn.chemutils.reaction_equation import (
    ReactionEquation,
    apply_to_compounds,
    merge_reactants_and_agents,
    sort_compounds,
)
from rxn.chemutils.reaction_smiles import (
    ReactionFormat,
    parse_any_reaction_smiles,
    to_reaction_smiles,
)

# rdBase.DisableLog("rdApp.error")
rdBase.DisableLog("rdApp.warning")

Molecule = Union[str, Chem.Mol]
SMI_REGEX_PATTERN = r"(\[[^\]]+]|Br?|Cl?|N|O|S|P|F|I|b|c|n|o|s|p|\(|\)|\.|=|#|-|\+|\\|\/|:|~|@|\?|>>?|\*|\$|\%[0-9]{2}|[0-9])"
BAD_TOKS = ["[CLS]", "[SEP]"]  # Default Bad Tokens


def get_rdkit_feature(
    mol: Molecule, radius: int = 2, num_bits: int = 2048, method="morgan_binary"
) -> np.ndarray:
    mol = Chem.MolFromSmiles(mol) if type(mol) == str else mol
    if method == "morgan_binary":
        features_vec = AllChem.GetMorganFingerprintAsBitVect(
            mol, radius, nBits=num_bits
        )
    elif method == "morgan_counts":
        features_vec = AllChem.GetHashedMorganFingerprint(mol, radius, nBits=num_bits)
    elif method == "rdkit_fingerprint":
        features_vec = Chem.RDKFingerprint(mol)
    else:
        raise NotImplementedError

    features = np.zeros((1,))
    DataStructs.ConvertToNumpyArray(features_vec, features)

    return features


def make_mol(s: str, keep_h: bool, add_h: bool):
    """
    From: https://github.com/chemprop/chemprop/blob/master/chemprop/rdkit.py

    Builds an RDKit molecule from a SMILES string.

    :param s: SMILES string.
    :param keep_h: Boolean whether to keep hydrogens in the input smiles. This does not add hydrogens, it only keeps them if they are specified.
    :return: RDKit molecule.
    """
    if keep_h:
        mol = Chem.MolFromSmiles(s, sanitize=False)
        Chem.SanitizeMol(
            mol,
            sanitizeOps=Chem.SanitizeFlags.SANITIZE_ALL
            ^ Chem.SanitizeFlags.SANITIZE_ADJUSTHS,
        )
    else:
        mol = Chem.MolFromSmiles(s)
    if add_h:
        mol = Chem.AddHs(mol)
    return mol


def generate_scaffold(
    mol: Union[str, Chem.Mol, Tuple[Chem.Mol, Chem.Mol]],
    include_chirality: bool = False,
) -> str:
    """
    From: https://github.com/chemprop/chemprop/blob/master/chemprop/data/scaffold.py

    Computes the Bemis-Murcko scaffold for a SMILES string.

    :param mol: A SMILES or an RDKit molecule.
    :param include_chirality: Whether to include chirality in the computed scaffold..
    :return: The Bemis-Murcko scaffold for the molecule.
    """
    if isinstance(mol, str):
        mol = make_mol(mol, keep_h=False, add_h=False)
    if isinstance(mol, tuple):
        mol = mol[0]
    scaffold = MurckoScaffold.MurckoScaffoldSmiles(
        mol=mol, includeChirality=include_chirality
    )

    return scaffold


def tokenize_smiles(smiles: str, return_as_str=False) -> Union[List[str], str]:
    """
    From https://github.com/rxn4chemistry/rxnmapper/blob/main/rxnmapper/smiles_utils.py

    Tokenize a SMILES molecule or reaction"""
    regex = re.compile(SMI_REGEX_PATTERN)
    tokens = [token for token in regex.findall(smiles)]
    assert smiles == "".join(tokens)
    if return_as_str:
        tokens = " ".join(tokens)
    return tokens


def get_atom_types(smiles: str):
    """
    From https://github.com/rxn4chemistry/rxnmapper/blob/main/rxnmapper/smiles_utils.py

    Return atomic numbers for every token in (reaction) SMILES"""
    atom_tokens = get_atom_tokens_mask(smiles)
    if ">>" in smiles:
        precursors, products = smiles.split(">>")

        atom_types = get_atom_types_smiles(precursors)
        atom_types += get_atom_types_smiles(products)

    else:
        atom_types = get_atom_types_smiles(smiles)

    token_types = []
    atom_counter = 0

    for i in atom_tokens:
        if i == 1:
            token_types.append(atom_types[atom_counter])
            atom_counter += 1
        else:
            token_types.append(-1)
    return token_types


def get_atom_types_smiles(smiles: str) -> List[int]:
    """
    From https://github.com/rxn4chemistry/rxnmapper/blob/main/rxnmapper/smiles_utils.py

    Convert each atom in a SMILES into a list of their atomic numbers

    Args:
        smiles: SMILES representation of a molecule or set of molecules.

    Returns:
        List of atom numbers for each atom in the smiles. Reports atoms in the same order they were passed in the original SMILES
    """
    # If `smiles` is a set of molecules, it may contain a "~".
    smiles_mol = smiles_to_mol(smiles.replace("~", "."), sanitize=False)

    atom_types = [atom.GetAtomicNum() for atom in smiles_mol.GetAtoms()]

    return atom_types


def is_atom(token: str, special_tokens: List[str] = BAD_TOKS) -> bool:
    """
    From https://github.com/rxn4chemistry/rxnmapper/blob/main/rxnmapper/smiles_utils.py

    Determine whether a token is an atom.

    Args:
        token: Token fed into the transformer model
        special_tokens: List of tokens to consider as non-atoms (often introduced by tokenizer)

    Returns:
        bool: True if atom, False if not
    """
    bad_toks = set(special_tokens)
    normal_atom = token[0].isalpha() or token[0] == "["
    is_bad = token in bad_toks
    return (not is_bad) and normal_atom


def number_tokens(tokens: List[str], special_tokens: List[str] = BAD_TOKS) -> List[int]:
    """
    From https://github.com/rxn4chemistry/rxnmapper/blob/main/rxnmapper/smiles_utils.py

    Map list of tokens to a list of numbered atoms

    Args:
        tokens: Tokenized SMILES
        special_tokens: List of tokens to not consider as atoms

    Example:
        >>> number_tokens(['[CLS]', 'C', '.', 'C', 'C', 'C', 'C', 'C', 'C','[SEP]'])
                #=> [-1, 0, -1, 1, 2, 3, 4, 5, 6, -1]
    """
    atom_num = 0
    isatm = partial(is_atom, special_tokens=special_tokens)

    def check_atom(t):
        if isatm(t):
            nonlocal atom_num
            ind = atom_num
            atom_num = atom_num + 1
            return ind
        return -1

    out = [check_atom(t) for t in tokens]

    return out


def get_graph_distance_matrix(smiles: str):
    """
    From https://github.com/rxn4chemistry/rxnmapper/blob/main/rxnmapper/smiles_utils.py

    Compute graph distance matrix between atoms. Only works for single molecules atm and not for rxns

    Args:
        smiles {[type]} -- [description]

    Returns:
        Numpy array representing the graphwise distance between each atom and every other atom in the molecular SMILES
    """
    mol = smiles_to_mol(smiles, sanitize=False)
    return Chem.GetDistanceMatrix(mol)


def get_adjacency_matrix(smiles: str):
    """
    From https://github.com/rxn4chemistry/rxnmapper/blob/main/rxnmapper/smiles_utils.py

    Compute adjacency matrix between atoms. Only works for single molecules atm and not for rxns

    Args:
        smiles: SMILES representation of a molecule

    Returns:
        Numpy array representing the adjacency between each atom and every other atom in the molecular SMILES.
        Equivalent to `distance_matrix[distance_matrix == 1]`
    """

    mol = smiles_to_mol(smiles, sanitize=False)
    return Chem.GetAdjacencyMatrix(mol)


def is_mol_end(a: str, b: str) -> bool:
    """
    From https://github.com/rxn4chemistry/rxnmapper/blob/main/rxnmapper/smiles_utils.py

    Determine if `a` and `b` are both tokens within a molecule (Used by the `group_with` function).

    Returns False whenever either `a` or `b` is a molecule delimeter (`.` or `>>`)"""
    no_dot = (a != ".") and (b != ".")
    no_arrow = (a != ">>") and (b != ">>")

    return no_dot and no_arrow


def group_with(predicate, xs: List[Any]):
    """
    From https://github.com/rxn4chemistry/rxnmapper/blob/main/rxnmapper/smiles_utils.py

    Takes a list and returns a list of lists where each sublist's elements are
    all satisfied pairwise comparison according to the provided function.
    Only adjacent elements are passed to the comparison function

        Original implementation here: https://github.com/slavaGanzin/ramda.py/blob/master/ramda/group_with.py

        Args:
            predicate ( f(a,b) => bool): A function that takes two subsequent inputs and returns True or Fale
            xs: List to group
    """
    out = []
    is_str = isinstance(xs, str)
    group = [xs[0]]

    for x in xs[1:]:
        if predicate(group[-1], x):
            group += [x]
        else:
            out.append("".join(group) if is_str else group)
            group = [x]

    out.append("".join(group) if is_str else group)

    return out


def split_into_mols(tokens: List[str]) -> List[List[str]]:
    """
    From https://github.com/rxn4chemistry/rxnmapper/blob/main/rxnmapper/smiles_utils.py

    Split a reaction SMILES into SMILES for each molecule"""
    split_toks = group_with(is_mol_end, tokens)
    return split_toks


def tokens_to_smiles(tokens: List[str], special_tokens: List[str]) -> str:
    """
    From https://github.com/rxn4chemistry/rxnmapper/blob/main/rxnmapper/smiles_utils.py

    Combine tokens into valid SMILES string, filtering out special tokens

    Args:
        tokens: Tokenized SMILES
        special_tokens: Tokens to not count as atoms

    Returns:
        SMILES representation of provided tokens, without the special tokens
    """
    bad_toks = set(special_tokens)
    return "".join([t for t in tokens if t not in bad_toks])


def tokens_to_adjacency(tokens: List[str]) -> np.array:
    """
    From https://github.com/rxn4chemistry/rxnmapper/blob/main/rxnmapper/smiles_utils.py

    Convert a tokenized reaction SMILES into a giant adjacency matrix.

    Note that this is a large, sparse Block Diagonal matrix of the adjacency matrix for each molecule in the reaction.

    Args:
        tokens: Tokenized SMILES representation

    Returns:
        Numpy Array, where non-zero entries in row `i` indicate the tokens that are atom-adjacent to token `i`
    """
    from scipy.linalg import block_diag

    mol_tokens = split_into_mols(tokens)

    smiles = [
        tokens_to_smiles(mol, BAD_TOKS) for mol in mol_tokens
    ]  # Cannot process SMILES if it is a '.' or '>>'

    # Calculate adjacency matrix for reaction
    altered_smiles = [
        s for s in smiles if s not in {".", "~", ">>"}
    ]  # Only care about atoms
    adjacency_mats = [
        get_adjacency_matrix(s) for s in altered_smiles
    ]  # Or filter if we don't need to save the spot
    rxn_mask = block_diag(*adjacency_mats)
    return rxn_mask


def get_mask_for_tokens(tokens: List[str], special_tokens: List[str] = []) -> List[int]:
    """
    From https://github.com/rxn4chemistry/rxnmapper/blob/main/rxnmapper/smiles_utils.py

    Return a mask for a tokenized smiles, where atom tokens
    are converted to 1 and other tokens to 0.

    e.g. c1ccncc1 would give [1, 0, 1, 1, 1, 1, 1, 0]

    Args:
        smiles: Smiles string of reaction
        special_tokens: Any special tokens to explicitly not call an atom. E.g. "[CLS]" or "[SEP]"

    Returns:
        Binary mask as a list where non-zero elements represent atoms
    """
    check_atom = partial(is_atom, special_tokens=special_tokens)

    atom_token_mask = [1 if check_atom(t) else 0 for t in tokens]
    return atom_token_mask


def tok_mask(tokens: List[str], special_tokens: List[str] = BAD_TOKS) -> np.array:
    """
    From https://github.com/rxn4chemistry/rxnmapper/blob/main/rxnmapper/smiles_utils.py

    Return a mask for a tokenized smiles, where atom tokens
    are converted to 1 and other tokens to 0.

    e.g. c1ccncc1 would give [1, 0, 1, 1, 1, 1, 1, 0]

    Args:
        smiles: Smiles string of reaction
        special_tokens: Any special tokens to explicitly not call an atom. E.g. "[CLS]" or "[SEP]"

    Returns:
        Binary mask as a boolean numpy array where True elements represent atoms
    """
    mask = get_mask_for_tokens(tokens, special_tokens=special_tokens)
    return np.array(mask).astype(np.bool)


def get_atom_tokens_mask(smiles: str, special_tokens: List[str] = []):
    """
    From https://github.com/rxn4chemistry/rxnmapper/blob/main/rxnmapper/smiles_utils.py

    Return a mask for a smiles, where atom tokens
    are converted to 1 and other tokens to 0.

    e.g. c1ccncc1 would give [1, 0, 1, 1, 1, 1, 1, 0]

    Args:
        smiles: Smiles string of reaction
        special_tokens: Any special tokens to explicitly not call an atom. E.g. "[CLS]" or "[SEP]"

    Returns:
        Binary mask as a list where non-zero elements represent atoms
    """
    tokens = tokenize_smiles(smiles)
    return get_mask_for_tokens(tokens, special_tokens)


def canonicalize_and_atom_map(smi: str, return_equivalent_atoms=False):
    """
    From https://github.com/rxn4chemistry/rxnmapper/blob/main/rxnmapper/smiles_utils.py

    Remove atom mapping, canonicalize and return mapping numbers in order of canonicalization.

    Args:
        smi: reaction SMILES str
        return_equivalent_atoms

    Returns:

    """
    mol = smiles_to_mol(smi, sanitize=False)
    for atom in mol.GetAtoms():
        if atom.HasProp("molAtomMapNumber"):
            atom_map = atom.GetAtomMapNum()
            atom.SetProp("atom_map", str(atom_map))
            atom.ClearProp("molAtomMapNumber")
        else:
            atom.SetProp("atom_map", str(0))
    can_smi = Chem.MolToSmiles(mol)
    order = list(
        mol.GetPropsAsDict(includePrivate=True, includeComputed=True)[
            "_smilesAtomOutputOrder"
        ]
    )

    atom_maps_canonical = [mol.GetAtoms()[idx].GetProp("atom_map") for idx in order]

    if not return_equivalent_atoms:
        return (can_smi, atom_maps_canonical)

    raise NotImplementedError


def generate_atom_mapped_reaction_atoms(
    rxn: str, product_atom_maps, expected_atom_maps=None, canonical: bool = False
) -> ReactionEquation:
    """
    From https://github.com/rxn4chemistry/rxnmapper/blob/main/rxnmapper/smiles_utils.py

    Generate atom-mapped reaction from unmapped reaction and
    product-2-reactant atoms mapping vector.
    Args:
        rxn: unmapped reaction, in the format that the transformer model relies on.
        product_atom_maps: product to reactant atom maps.
        expected_atom_maps: if given, return the differences.
        canonical: whether to canonicalize the resulting SMILES.

    Returns: Atom-mapped reaction

    """

    reactants, agents, products = parse_any_reaction_smiles(rxn)
    precursors_mols = [smiles_to_mol(pr, sanitize=False) for pr in reactants + agents]
    products_mols = [smiles_to_mol(prod, sanitize=False) for prod in products]

    precursors_atom_maps = []

    differing_maps = []

    product_mapping_dict = {}

    i = -1
    atom_mapped_precursors_list = []
    for precursor_mol in precursors_mols:
        for atom in precursor_mol.GetAtoms():
            i += 1
            if i in product_atom_maps:
                # atom maps start at an index of 1
                corresponding_product_atom_map = product_atom_maps.index(i) + 1
                precursors_atom_maps.append(corresponding_product_atom_map)
                atom.SetProp("molAtomMapNumber", str(corresponding_product_atom_map))

                indices = [idx for idx, x in enumerate(product_atom_maps) if x == i]

                if len(indices) > 1:
                    for idx in indices[1:]:
                        product_mapping_dict[idx] = corresponding_product_atom_map

                if expected_atom_maps is not None:
                    if (
                        i not in expected_atom_maps
                        or corresponding_product_atom_map
                        != expected_atom_maps.index(i) + 1
                    ):
                        differing_maps.append(corresponding_product_atom_map)
        atom_mapped_precursors_list.append(
            Chem.MolToSmiles(precursor_mol, canonical=canonical)
        )

    i = -1
    atom_mapped_products_list = []
    for products_mol in products_mols:
        for atom in products_mol.GetAtoms():
            i += 1
            atom_map = product_mapping_dict.get(i, i + 1)
            atom.SetProp("molAtomMapNumber", str(atom_map))
        atom_mapped_products_list.append(
            Chem.MolToSmiles(products_mol, canonical=canonical)
        )

    atom_mapped_rxn = ReactionEquation(
        atom_mapped_precursors_list, [], atom_mapped_products_list
    )

    if expected_atom_maps is not None:
        return atom_mapped_rxn, differing_maps

    return atom_mapped_rxn


def canonicalize_smi(smi: str, remove_mapping: bool = False) -> str:
    """
    From https://github.com/rxn4chemistry/rxnmapper/blob/main/rxnmapper/smiles_utils.py

    Convert a SMILES string into its canonicalized form

    Args:
        smi: Reaction SMILES
        remove_mapping: If True, remove atom mapping information from the canonicalized SMILES output

    Raises:
        InvalidSmiles: if the SMILES string cannot be canonicalized.

    Returns:
        SMILES reaction, canonicalized, as a string
    """
    if remove_mapping:
        smi = remove_atom_mapping(smi)

    return canonicalize_smiles(smi)


def process_reaction(reaction: ReactionEquation) -> ReactionEquation:
    """
    From https://github.com/rxn4chemistry/rxnmapper/blob/main/rxnmapper/smiles_utils.py

    Remove atom-mapping, move reagents to reactants and canonicalize reaction.

    Args:
        reaction: Reaction equation to process.

    Returns:
        Processed reaction.
    """
    reaction = merge_reactants_and_agents(reaction)

    try:
        canonicalize_and_remove_atom_map = partial(
            canonicalize_smi, remove_mapping=True
        )
        reaction = apply_to_compounds(reaction, canonicalize_and_remove_atom_map)
    except InvalidSmiles:
        return ReactionEquation([], [], [])

    reaction = sort_compounds(reaction)
    return reaction


def process_reaction_with_product_maps_atoms(rxn, skip_if_not_in_precursors=False):
    """
    From https://github.com/rxn4chemistry/rxnmapper/blob/main/rxnmapper/smiles_utils.py

    Remove atom-mapping, move reagents to reactants and canonicalize reaction.
    If fragment group information is given, keep the groups together using
    the character defined with fragment_bond.

    Args:
        rxn: Reaction SMILES
        skip_if_not_in_precursors: accept unmapped atoms in the product (default: False)

    Returns: joined_precursors>>joined_products reaction SMILES
    """
    reactants, reagents, products = rxn.split(">")
    try:
        precursors = [canonicalize_and_atom_map(r) for r in reactants.split(".")]
        if len(reagents) > 0:
            precursors += [canonicalize_and_atom_map(r) for r in reagents.split(".")]
        products = [canonicalize_and_atom_map(p) for p in products.split(".")]
    except InvalidSmiles:
        return ""
    sorted_precursors = sorted(precursors, key=lambda x: x[0])
    sorted_products = sorted(products, key=lambda x: x[0])
    joined_precursors = ".".join([p[0] for p in sorted_precursors])
    joined_products = ".".join([p[0] for p in sorted_products])
    precursors_atom_maps = [
        i for p in sorted_precursors for i in p[1]
    ]  # could contain duplicate entries
    product_atom_maps = [
        i for p in sorted_products for i in p[1]
    ]  # could contain duplicate entries

    joined_rxn = f"{joined_precursors}>>{joined_products}"

    products_maps = []
    warnings_list = []

    for p_map in product_atom_maps:
        if skip_if_not_in_precursors and p_map not in precursors_atom_maps:
            products_maps.append(-1)
        elif int(p_map) == 0:
            products_maps.append(-1)
        else:
            corresponding_precursors_atom = precursors_atom_maps.index(p_map)
            if (
                corresponding_precursors_atom in products_maps
            ):  # handle equivalent atoms
                found_alternative = False
                for atom_idx, precursor_map in enumerate(precursors_atom_maps):
                    if (precursor_map == p_map) and atom_idx not in products_maps:
                        products_maps.append(atom_idx)
                        found_alternative = True
                        break
                if not found_alternative:
                    warnings_list.append(
                        f"Two product atoms mapped to the same precursor atom: {rxn}"
                    )
                    products_maps.append(corresponding_precursors_atom)
            else:
                products_maps.append(corresponding_precursors_atom)
    for w in list(set(warnings_list)):
        warnings.warn(w)
    return joined_rxn, products_maps


def standardize_reaction(reaction: str):
    # reaction_formats = determine_format(reaction)
    reaction = parse_any_reaction_smiles(reaction)
    reaction = process_reaction(reaction)
    reaction = to_reaction_smiles(
        reaction, reaction_format=ReactionFormat.STANDARD_WITH_TILDE
    )
    return reaction

    # encoded_ids = self.tokenizer.batch_encode_plus(
    #     rxn_smiles_list,
    #     padding=True,
    #     return_tensors="pt",
    # )
    # parsed_input = {k: v.to(self.device) for k, v in encoded_ids.items()}
    # return encoded_ids


def remove_atom_maps_manual(smiles: str):
    """Remove atom maps from smiles string using RegEx"""
    smi = re.sub(":[0-9]+", "", smiles)
    return smi


def remove_atom_maps(smiles: str, sanitize=True):
    """Remove atom maps from smiles string"""
    mol = Chem.MolFromSmiles(smiles, sanitize=sanitize)
    for atom in mol.GetAtoms():
        atom.SetAtomMapNum(0)
    return Chem.MolToSmiles(mol)


def assign_dummy_atom_maps(smiles: str):
    """Assign atom numbers to smiles string for use in graph editing"""
    mol = Chem.MolFromSmiles(smiles)
    for i, atom in enumerate(mol.GetAtoms()):
        atom.SetAtomMapNum(i + 1)
    return Chem.MolToSmiles(mol)



from typing import Dict
from clipzyme.utils.registry import register_object
from clipzyme.utils.classes import Nox
import torchmetrics
from torchmetrics import Metric
from torchmetrics.utilities.compute import auc as compute_auc
import torch


@register_object("classification", "metric")
class BaseClassification(Metric, Nox):
    def __init__(self, args) -> None:
        """
        Computes standard classification metrics

        Args:
            predictions_dict: dictionary obtained from computing loss and model outputs
                * should contain the keys ['probs', 'preds', 'golds']
            args: argparser Namespace

        Returns:
            stats_dict (dict): contains (where applicable) values for accuracy, confusion matrix, precision, recall, f1, precision-recall auc, roc auc

        Note:
            Binary: two labels
            Multiclass: more than two labels
            Multilabel: potentially more than one label per sample (independent classes)
        """
        super().__init__()
        self.task_type = args.task_type
        if args.task_type != "multilabel":
            self.accuracy_metric = torchmetrics.Accuracy(
                task=args.task_type,
                num_classes=args.num_classes,
            )
            self.auroc_metric = torchmetrics.AUROC(
                task=args.task_type, num_classes=args.num_classes
            )
            self.f1_metric = torchmetrics.F1Score(
                task=args.task_type,
                num_classes=args.num_classes,
            )
            self.macro_f1_metric = torchmetrics.F1Score(
                task=args.task_type,
                num_classes=args.num_classes,
                average="macro",
            )
            self.ap_metric = torchmetrics.AveragePrecision(
                task=args.task_type, num_classes=args.num_classes
            )
            self.auprc_metric = torchmetrics.PrecisionRecallCurve(
                task=args.task_type, num_classes=args.num_classes
            )
            self.precision_metric = torchmetrics.Precision(
                task=args.task_type,
                num_classes=args.num_classes,
            )
            self.recall_metric = torchmetrics.Recall(
                task=args.task_type,
                num_classes=args.num_classes,
            )
        elif self.task_type == "multilabel":
            self.accuracy_metric = torchmetrics.Accuracy(
                task=args.task_type,
                num_labels=args.num_classes,
            )
            self.auroc_metric = torchmetrics.AUROC(
                task=args.task_type, num_labels=args.num_classes
            )
            self.f1_metric = torchmetrics.F1Score(
                task=args.task_type,
                num_labels=args.num_classes,
            )
            self.macro_f1_metric = torchmetrics.F1Score(
                task=args.task_type,
                num_labels=args.num_classes,
                average="macro",
            )
            self.ap_metric = torchmetrics.AveragePrecision(
                task=args.task_type, num_labels=args.num_classes
            )
            self.auprc_metric = torchmetrics.PrecisionRecallCurve(
                task=args.task_type, num_labels=args.num_classes
            )
            self.precision_metric = torchmetrics.Precision(
                task=args.task_type,
                num_labels=args.num_classes,
            )
            self.recall_metric = torchmetrics.Recall(
                task=args.task_type,
                num_labels=args.num_classes,
            )
        else:
            self.accuracy_metric = torchmetrics.Accuracy(
                task=args.task_type,
            )
            self.auroc_metric = torchmetrics.AUROC(
                task=args.task_type,
            )
            self.f1_metric = torchmetrics.F1Score(
                task=args.task_type,
            )
            self.macro_f1_metric = torchmetrics.F1Score(
                task=args.task_type,
                average="macro",
            )
            self.ap_metric = torchmetrics.AveragePrecision(
                task=args.task_type,
            )
            self.auprc_metric = torchmetrics.PrecisionRecallCurve(
                task=args.task_type,
            )
            self.precision_metric = torchmetrics.Precision(
                task=args.task_type,
            )
            self.recall_metric = torchmetrics.Recall(
                task=args.task_type,
            )

    @property
    def metric_keys(self):
        return ["probs", "preds", "golds"]

    def update(self, predictions_dict, args) -> Dict:
        probs = predictions_dict["probs"]  # B, C (float)
        preds = predictions_dict["preds"]  # B
        golds = predictions_dict["golds"].int()  # B

        if self.task_type == "binary":
            if len(preds.shape) >= 2:
                preds = preds[:, 1]
            if len(probs.shape) >= 2:
                probs = probs[:, 1]

        self.accuracy_metric.update(preds, golds)
        self.auroc_metric.update(probs, golds)
        self.auprc_metric.update(probs, golds)
        self.f1_metric.update(probs, golds)
        self.macro_f1_metric.update(probs, golds)
        self.precision_metric.update(probs, golds)
        self.recall_metric.update(probs, golds)
        self.ap_metric.update(probs, golds)

    def compute(self) -> Dict:
        pr, rc, _ = self.auprc_metric.compute()
        if self.task_type != "binary":  # list per class or per label if not binary task
            pr_auc = [compute_auc(rc_i, pr_i) for rc_i, pr_i in zip(rc, pr)]
            pr_auc = torch.mean(torch.stack(pr_auc))
        else:
            pr_auc = compute_auc(rc, pr)
        stats_dict = {
            "accuracy": self.accuracy_metric.compute(),
            "roc_auc": self.auroc_metric.compute(),
            "pr_auc": pr_auc,
            "f1": self.f1_metric.compute(),
            "macro_f1": self.macro_f1_metric.compute(),
            "precision": self.precision_metric.compute(),
            "recall": self.recall_metric.compute(),
            "average_precision": self.ap_metric.compute(),
        }
        return stats_dict

    def reset(self):
        super().reset()
        self.accuracy_metric.reset()
        self.auroc_metric.reset()
        self.auprc_metric.reset()
        self.f1_metric.reset()
        self.macro_f1_metric.reset()
        self.precision_metric.reset()
        self.recall_metric.reset()
        self.ap_metric.reset()

    @staticmethod
    def add_args(parser) -> None:
        """Add class specific args

        Args:
            parser (argparse.ArgumentParser): argument parser
        """
        parser.add_argument(
            "--task_type",
            type=str,
            default=None,
            required=True,
            choices=["binary", "multiclass", "multilabel"],
            help="type of task",
        )


@register_object("dummy_classification", "metric")
class DummyBaseClassification(Metric, Nox):
    def __init__(self, args) -> None:
        """
        Computes standard classification metrics

        Args:
            predictions_dict: dictionary obtained from computing loss and model outputs
                * should contain the keys ['probs', 'preds', 'golds']
            args: argparser Namespace

        Returns:
            stats_dict (dict): contains (where applicable) values for accuracy, confusion matrix, precision, recall, f1, precision-recall auc, roc auc

        Note:
            Binary: two labels
            Multiclass: more than two labels
            Multilabel: potentially more than one label per sample (independent classes)
        """
        super().__init__()
        self.task_type = args.task_type

        self.accuracy_metric = torchmetrics.Accuracy(
            task=args.task_type,
            num_classes=args.num_classes,
        )
        self.auroc_metric = torchmetrics.AUROC(
            task=args.task_type, num_classes=args.num_classes
        )
        self.f1_metric = torchmetrics.F1Score(
            task=args.task_type,
            num_classes=args.num_classes,
        )
        self.macro_f1_metric = torchmetrics.F1Score(
            task=args.task_type,
            num_classes=args.num_classes,
            average="macro",
        )
        self.ap_metric = torchmetrics.AveragePrecision(
            task=args.task_type, num_classes=args.num_classes
        )
        self.auprc_metric = torchmetrics.PrecisionRecallCurve(
            task=args.task_type, num_classes=args.num_classes
        )
        self.precision_metric = torchmetrics.Precision(
            task=args.task_type,
            num_classes=args.num_classes,
        )
        self.recall_metric = torchmetrics.Recall(
            task=args.task_type,
            num_classes=args.num_classes,
        )

    @property
    def metric_keys(self):
        return ["y", "probs", "preds", "golds"]

    def update(self, predictions_dict, args) -> Dict:
        probs = predictions_dict["probs"]  # B, C (float)
        probs = torch.diagonal(probs)  # I am only looking at samples I added
        preds = predictions_dict["preds"]  # B
        preds = torch.diagonal(preds)
        golds = predictions_dict["y"].int()  # B

        self.accuracy_metric.update(preds, golds)
        self.auroc_metric.update(probs, golds)
        self.auprc_metric.update(probs, golds)
        self.f1_metric.update(probs, golds)
        self.macro_f1_metric.update(probs, golds)
        self.precision_metric.update(probs, golds)
        self.recall_metric.update(probs, golds)
        self.ap_metric.update(probs, golds)

    def compute(self) -> Dict:
        pr, rc, _ = self.auprc_metric.compute()
        if self.task_type != "binary":  # list per class or per label if not binary task
            pr_auc = [compute_auc(rc_i, pr_i) for rc_i, pr_i in zip(rc, pr)]
            pr_auc = torch.mean(torch.stack(pr_auc))
        else:
            pr_auc = compute_auc(rc, pr)
        stats_dict = {
            "accuracy": self.accuracy_metric.compute(),
            "roc_auc": self.auroc_metric.compute(),
            "pr_auc": pr_auc,
            "f1": self.f1_metric.compute(),
            "macro_f1": self.macro_f1_metric.compute(),
            "precision": self.precision_metric.compute(),
            "recall": self.recall_metric.compute(),
            "average_precision": self.ap_metric.compute(),
        }
        return stats_dict

    def reset(self):
        self.accuracy_metric.reset()
        self.auroc_metric.reset()
        self.auprc_metric.reset()
        self.f1_metric.reset()
        self.macro_f1_metric.reset()
        self.precision_metric.reset()
        self.recall_metric.reset()
        self.ap_metric.reset()

    @staticmethod
    def add_args(parser) -> None:
        """Add class specific args

        Args:
            parser (argparse.ArgumentParser): argument parser
        """
        parser.add_argument(
            "--task_type",
            type=str,
            default=None,
            required=True,
            choices=["binary", "multiclass", "multilabel"],
            help="type of task",
        )


@register_object("seq2seq_classification", "metric")
class Seq2SeqClassification(Metric, Nox):
    """Same as BaseClassification but samplewise multidim_average"""

    def __init__(self, args) -> None:
        super().__init__()
        self.ignore_index = args.ignore_index
        self.task_type = args.task_type
        self.accuracy_metric = torchmetrics.Accuracy(
            task=args.task_type,
            num_classes=args.num_classes,
            num_labels=args.num_classes,
            multidim_average="samplewise",
            ignore_index=args.ignore_index,
        )
        self.f1_metric = torchmetrics.F1Score(
            task=args.task_type,
            num_classes=args.num_classes,
            num_labels=args.num_classes,
            multidim_average="samplewise",
            ignore_index=args.ignore_index,
        )
        self.macro_f1_metric = torchmetrics.F1Score(
            task=args.task_type,
            num_labels=args.num_classes,
            num_classes=args.num_classes,
            average="macro",
            multidim_average="samplewise",
            ignore_index=args.ignore_index,
        )
        self.precision_metric = torchmetrics.Precision(
            task=args.task_type,
            num_labels=args.num_classes,
            num_classes=args.num_classes,
            multidim_average="samplewise",
            ignore_index=args.ignore_index,
        )
        self.recall_metric = torchmetrics.Recall(
            task=args.task_type,
            num_labels=args.num_classes,
            num_classes=args.num_classes,
            multidim_average="samplewise",
            ignore_index=args.ignore_index,
        )
        self.top_1_metric = TopK(
            task=args.task_type,
            num_labels=args.num_classes,
            num_classes=args.num_classes,
            ignore_index=args.ignore_index,
        ).to(self.accuracy_metric.device)
        # don't have multidim_average arg
        self.auroc_metric = torchmetrics.AUROC(
            task=args.task_type,
            num_labels=args.num_classes,
            num_classes=args.num_classes,
            ignore_index=args.ignore_index,
        )
        self.ap_metric = torchmetrics.AveragePrecision(
            task=args.task_type,
            num_labels=args.num_classes,
            num_classes=args.num_classes,
            ignore_index=args.ignore_index,
        )
        self.auprc_metric = torchmetrics.PrecisionRecallCurve(
            task=args.task_type,
            num_labels=args.num_classes,
            num_classes=args.num_classes,
            ignore_index=args.ignore_index,
        )

    @property
    def metric_keys(self):
        return ["probs", "preds", "golds", "preds_mask"]

    def update(self, predictions_dict, args) -> Dict:
        probs = predictions_dict["probs"]  # B, N, C (float)
        preds = predictions_dict["preds"]  # B, N
        golds = predictions_dict["golds"].long()  # B, N
        preds_mask = predictions_dict["preds_mask"]

        if self.ignore_index is not None:
            preds[preds_mask == self.ignore_index] = self.ignore_index

        self.accuracy_metric.update(preds, golds)
        self.f1_metric.update(preds, golds)
        self.macro_f1_metric.update(preds, golds)
        self.precision_metric.update(preds, golds)
        self.recall_metric.update(preds, golds)
        self.top_1_metric.update(preds, golds, preds_mask)

        # for sample_prob, sample_gold in zip(probs, golds):
        #     self.auroc_metric.update(sample_prob, sample_gold)
        #     self.ap_metric.update(sample_prob, sample_gold)
        #     self.auprc_metric.update(sample_prob, sample_gold)

    def compute(self) -> Dict:
        # pr, rc, _ = self.auprc_metric.compute()
        # if self.task_type != "binary":  # list per class or per label if not binary task
        #     pr_auc = [compute_auc(rc_i, pr_i) for rc_i, pr_i in zip(rc, pr)]
        #     pr_auc = torch.mean(torch.stack(pr_auc))
        # else:
        #     pr_auc = compute_auc(rc, pr)
        stats_dict = {
            "accuracy": self.accuracy_metric.compute().mean(),
            # "roc_auc": self.auroc_metric.compute(),
            # pr_auc": pr_auc,
            "f1": self.f1_metric.compute().mean(),
            "macro_f1": self.macro_f1_metric.compute().mean(),
            "precision": self.precision_metric.compute().mean(),
            "recall": self.recall_metric.compute().mean(),
            "top_1": self.top_1_metric.compute(),
        }
        return stats_dict

    def reset(self):
        super().reset()
        self.accuracy_metric.reset()
        # self.auroc_metric.reset()
        # self.auprc_metric.reset()
        self.f1_metric.reset()
        self.macro_f1_metric.reset()
        self.precision_metric.reset()
        self.recall_metric.reset()
        self.ap_metric.reset()
        self.top_1_metric.reset()

    @staticmethod
    def add_args(parser) -> None:
        """Add class specific args

        Args:
            parser (argparse.ArgumentParser): argument parser
        """
        parser.add_argument(
            "--task_type",
            type=str,
            default=None,
            required=True,
            choices=["binary", "multiclass", "multilabel"],
            help="type of task",
        )
        parser.add_argument(
            "--ignore_index",
            type=int,
            default=-100,
            help="type of task",
        )


@register_object("regression", "metric")
class BaseRegression(Metric, Nox):
    def __init__(self, args) -> None:
        super().__init__()

        self.num_classes = args.num_classes
        self.mae = torchmetrics.MeanAbsoluteError()
        self.mse = torchmetrics.MeanSquaredError()
        self.pearson = torchmetrics.PearsonCorrCoef(num_outputs=args.num_classes)
        self.spearman = torchmetrics.SpearmanCorrCoef(num_outputs=args.num_classes)
        self.r2 = torchmetrics.R2Score(
            num_outputs=args.num_classes, multioutput=args.r2_multioutput
        )
        self.cosine_similarity = torchmetrics.CosineSimilarity(reduction="mean")

    @property
    def metric_keys(self):
        return ["probs", "golds"]

    def update(self, predictions_dict, args) -> Dict:
        probs = predictions_dict["probs"]  # B, C (float)
        golds = predictions_dict["golds"]  # B

        if probs.shape[-1] == 1:
            probs = probs.view(-1)
            golds = golds.view(-1)

        self.mae.update(probs, golds)
        self.mse.update(probs, golds)
        self.pearson.update(probs, golds)
        self.spearman.update(probs, golds)
        self.r2.update(probs, golds)
        if self.num_classes > 1:
            self.cosine_similarity.update(probs, golds)

    def compute(self) -> Dict:
        stats_dict = {
            "mae": self.mae.compute(),
            "mse": self.mse.compute(),
            "pearson": self.pearson.compute(),
            "spearman": self.spearman.compute(),
            "r2": self.r2.compute(),
        }
        if self.num_classes > 1:
            stats_dict["cosine_similarity"] = self.cosine_similarity.compute()
        return stats_dict

    def reset(self):
        super().reset()
        self.mae.reset()
        self.mse.reset()
        self.pearson.reset()
        self.spearman.reset()
        self.r2.reset()
        self.cosine_similarity.reset()

    @staticmethod
    def add_args(parser) -> None:
        """Add class specific args

        Args:
            parser (argparse.ArgumentParser): argument parser
        """
        parser.add_argument(
            "--r2_multioutput",
            type=str,
            default="uniform_average",
            choices=["raw_values", "uniform_average", "variance_weighted"],
            help="aggregation in the case of multiple output scores",
        )


class TopK(Metric):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        higher_is_better: Optional[bool] = True
        is_differentiable = False
        full_state_update: bool = False

        self.ignore_index = kwargs["ignore_index"]
        self.task = kwargs["task"]
        self.num_classes = kwargs["num_classes"]

        self.add_state("correct", default=torch.tensor(0), dist_reduce_fx="sum")
        self.add_state("total", default=torch.tensor(0), dist_reduce_fx="sum")

    def update(
        self, preds: torch.Tensor, target: torch.Tensor, preds_mask: torch.Tensor
    ):
        assert preds.shape == target.shape

        if self.ignore_index is not None:
            preds[preds_mask == self.ignore_index] = self.ignore_index

        correct = (preds == target).sum(1) == preds.shape[1]
        correct = correct.sum()
        total = torch.tensor(preds.shape[0])

        self.correct += correct.to(self.device)
        self.total += total.to(self.device)

    def compute(self):
        return self.correct.float() / self.total



import torch.nn as nn
from clipzyme.utils.classes import Nox
from abc import ABCMeta, abstractmethod

# from efficientnet_pytorch import EfficientNet
import math


class AbstractModel(nn.Module, Nox):
    __metaclass__ = ABCMeta

    def __init__(self):
        super(AbstractModel, self).__init__()



import os
from typing import List, Union
from rich import print
from tqdm import tqdm
import copy
import pandas as pd
import Bio
import Bio.PDB
from Bio.Data.IUPACData import protein_letters_3to1
import torch
from torch.utils import data
from torch_geometric.data import Data as pygData
from esm import pretrained
from clipzyme.utils.registry import register_object
from clipzyme.utils.wln_processing import get_bond_changes
from clipzyme.utils.pyg import from_mapped_smiles
from clipzyme.utils.protein_utils import (
    read_structure_file,
    filter_resolution,
    build_graph,
    compute_graph_edges,
    compute_node_embedding,
)


protein_letters_3to1.update({k.upper(): v for k, v in protein_letters_3to1.items()})


@register_object("reactions_dataset", "dataset")
class ReactionDataset(data.Dataset):
    def __init__(self, args) -> None:
        """
        Create a dataset of reactions and proteins from a CSV file

        Parameters
        ----------
        args: argparse.Namespace

        Raises
        ------
        ValueError
            If CSV file does not have headers ['reaction', 'sequence', 'protein_id', 'cif']
        """
        super(ReactionDataset, self).__init__()
        # check if csv file has correct headers
        csv_path = args.dataset_file_path
        esm_dir = args.esm_dir
        protein_cache_dir = args.protein_cache_dir

        with open(csv_path, "r") as f:
            headers = f.readline().strip().split(",")
            if (
                "reaction" not in headers
                or "protein_id" not in headers
                or "sequence" not in headers
                or "cif" not in headers
            ):
                raise ValueError(
                    "CSV file must have headers 'reaction', 'sequence', 'protein_id', and 'cif'"
                )

        print("Loading ESM model")
        esm_path = os.path.join(esm_dir, "esm2_t33_650M_UR50D.pt")
        model, alphabet = pretrained.load_model_and_alphabet(esm_path)
        self.esm_model = model
        self.alphabet = alphabet
        self.batch_converter = alphabet.get_batch_converter()

        print("Preparing dataset")
        csv_dataset = pd.read_csv(csv_path)
        self.dataset = self.create_dataset(csv_dataset)

        self.protein_cache_dir = protein_cache_dir

        print(self.SUMMARY_STATEMENT)

    def create_dataset(self, csv_dataset: pd.DataFrame) -> List[dict]:
        """
        Create dataset of reactions and proteins from CSV file

        Parameters
        ----------
        csv_path : str
            Path to CSV file with headers ['reaction', 'sequence', 'cif']

        Returns
        -------
        List[dict]
            _description_
        """

        dataset = []

        for rowid, row in tqdm(
            csv_dataset.iterrows(),
            desc="Building dataset",
            total=len(csv_dataset),
            ncols=100,
        ):
            reactants, products = row["reaction"].split(">>")
            reactants = reactants.split(".")
            products = products.split(".")
            sample = {
                "protein_id": row["protein_id"],
                "sequence": row["sequence"],
                "reaction": row["reaction"],
                "reactants": reactants,
                "products": products,
                "cif_path": row["cif"],
                "sample_id": f"sample_{rowid}",
            }
            # get bond changes
            try:
                sample["bond_changes"] = get_bond_changes(sample["reaction"])
            except Exception as e:
                sample["bond_changes"] = []

            if self.skip_sample(sample):
                continue

            # add reaction sample to dataset
            dataset.append(sample)

        return dataset

    def skip_sample(self, sample: dict) -> bool:
        """
        Skip sample if criteria are not met

        Parameters
        ----------
        sample : dict
            Sample dictionary from dataset

        Returns
        -------
        bool
            True if sample should be skipped
        """
        # if sequence is unknown
        sequence = sample["sequence"]
        if len(sequence) == 0:
            return True

        if len(sequence) > 650:
            return True

        # check if cif file exists
        if not os.path.exists(sample["cif_path"]):
            return True

        # if no bond changes
        if len(sample["bond_changes"]) == 0:
            return True

        return False

    def create_protein_graph(self, sample: dict) -> Union[pygData, None]:
        """
        Create pyg protein graph from CIF file

        Parameters
        ----------
        sample : dict
            dataset sample

        Returns
        -------
        data
            pygData object with protein graph
        """
        try:
            raw_path = sample["cif_path"]
            sample_id = sample["sample_id"]
            protein_parser = Bio.PDB.MMCIFParser()
            protein_resolution = "residue"
            graph_edge_args = {"knn_size": 10}
            center_protein = True

            # parse pdb
            all_res, all_atom, all_pos = read_structure_file(
                protein_parser, raw_path, sample_id
            )
            # filter resolution of protein (backbone, atomic, etc.)
            atom_names, seq, pos = filter_resolution(
                all_res,
                all_atom,
                all_pos,
                protein_resolution=protein_resolution,
            )
            # generate graph
            data = build_graph(atom_names, seq, pos, sample_id)
            # kNN graph
            data = compute_graph_edges(data, **graph_edge_args)
            if center_protein:
                center = data["receptor"].pos.mean(dim=0, keepdim=True)
                data["receptor"].pos = data["receptor"].pos - center
                data.center = center

            sequence = sample["sequence"]
            data.structure_sequence = sequence

            node_embeddings_args = {
                "model": self.esm_model,
                "model_location": self.esm_dir,
                "alphabet": self.alphabet,
                "batch_converter": self.batch_converter,
            }
            node_embedding = compute_node_embedding(data, **node_embeddings_args)
            # Fix sequence length mismatches
            if len(data["receptor"].seq) != node_embedding.shape[0]:
                print("Computing seq embedding for mismatched seq length")
                AA_seq = ""
                for char in seq:
                    AA_seq += protein_letters_3to1[char]

                data.structure_sequence = AA_seq
                data["receptor"].x = compute_node_embedding(
                    data, **node_embeddings_args
                )
            else:
                data["receptor"].x = node_embedding

            if len(data["receptor"].seq) != data["receptor"].x.shape[0]:
                return None

            if hasattr(data, "x") and not hasattr(data["receptor"], "x"):
                data["receptor"].x = data.x

            if not hasattr(data, "structure_sequence"):
                data.structure_sequence = "".join(
                    [protein_letters_3to1[char] for char in data["receptor"].seq]
                )

            keep_keys = {
                "receptor",
                "structure_sequence",
                ("receptor", "contact", "receptor"),
            }

            data_keys = data.to_dict().keys()
            for d_key in data_keys:
                if d_key not in keep_keys:
                    delattr(data, d_key)

            coors = data["receptor"].pos
            feats = data["receptor"].x
            edge_index = data["receptor", "contact", "receptor"].edge_index
            assert (
                coors.shape[0] == feats.shape[0]
            ), f"Number of nodes do not match between coors ({coors.shape[0]}) and feats ({feats.shape[0]})"

            assert (
                max(edge_index[0]) < coors.shape[0]
                and max(edge_index[1]) < coors.shape[0]
            ), "Edge index contains node indices not present in coors"

            return data

        except Exception as e:
            print(
                f"Could not create protein graph for:  {sample['protein_id']} because of the exception {e}"
            )
            return None

    def __len__(self):
        return len(self.dataset)

    def __getitem__(self, index):
        sample = self.dataset[index]

        try:
            reactants, products = (
                copy.deepcopy(sample["reactants"]),
                copy.deepcopy(sample["products"]),
            )
            protein_id = sample["protein_id"]

            reactants, atom_map2new_index = from_mapped_smiles(
                ".".join(reactants),
                encode_no_edge=True,
                use_one_hot_encoding=self.args.use_one_hot_mol_features,
            )
            products, _ = from_mapped_smiles(
                ".".join(products),
                encode_no_edge=True,
                use_one_hot_encoding=self.args.use_one_hot_mol_features,
            )

            bond_changes = [
                (atom_map2new_index[int(u)], atom_map2new_index[int(v)], btype)
                for u, v, btype in sample["bond_changes"]
            ]
            bond_changes = [(min(x, y), max(x, y), t) for x, y, t in bond_changes]
            reactants.bond_changes = bond_changes
            sample_id = sample["sample_id"]

            item = {
                "reactants": reactants,
                "products": products,
                "protein_id": protein_id,
                "sample_id": sample_id,
                "cif_path": sample["cif_path"],
                "sequence": sample["sequence"],
            }

            if self.protein_cache_dir:
                graph_path_cache = os.path.join(
                    self.protein_cache_dir, f"{protein_id}.pt"
                )
                try:
                    data = torch.load(graph_path_cache)
                    if data is None:
                        data = self.create_protein_graph(item)
                        torch.save(data, graph_path_cache)
                except:
                    data = self.create_protein_graph(item)
                    torch.save(data, graph_path_cache)
            else:
                data = self.create_protein_graph(item)

                item["graph"] = data

            return item

        except Exception as e:
            print(f"Could not load sample {sample_id} because of an exception {e}")

    @property
    def SUMMARY_STATEMENT(self) -> None:
        reactions = [d["reaction"] for d in self.dataset]
        proteins = [d["sequence"] for d in self.dataset]

        statement = f""" 
        DATASET CREATED:
        * Number of samples: {len(self.dataset)}
        * Number of reactions: {len(set(reactions))}
        * Number of proteins: {len(set(proteins))}
        """
        return statement



from clipzyme.utils.registry import register_object
import pytorch_lightning as pl
import os
from clipzyme.utils.classes import Nox
import wandb
import rdkit


@register_object("wandb", "logger")
class WandB(pl.loggers.WandbLogger, Nox):
    def __init__(self, args) -> None:
        super().__init__(
            project=args.project_name,
            name=args.experiment_name,
            entity=args.workspace,
            tags=args.logger_tags,
        )

    def setup(self, **kwargs):
        # "gradients", "parameters", "all", or None
        # # change "log_freq" log frequency of gradients and parameters (100 steps by default)
        if kwargs["args"].local_rank == 0:
            self.watch(kwargs["model"], log="all")
            self.experiment.config.update(kwargs["args"])

    def log_image(self, image, name):
        self.log_image(images=[image], caption=[name])

    def log_table(self, table_columns, table_data, data_types, table_name):
        """
        table_columns: list of column names
        table_data: list per column
        data_types: list of data type per column
        """
        formatted_data = []
        for datalist, datatype in zip(table_data, data_types):
            if datatype == "image":
                formatted_data.append([wandb.Image(dataitem) for dataitem in datalist])
            elif datatype == "video":
                formatted_data.append(
                    [
                        wandb.Video(dataitem, format="gif", fps=4)
                        for dataitem in datalist
                    ]
                )
            elif datatype == "smiles":
                fd = []
                for dataitem in datalist:
                    mol = rdkit.Chem.Draw.MolToImage(
                        rdkit.Chem.MolFromSmiles(dataitem), size=(200, 200)
                    )
                    item = wandb.Image(mol)
                    fd.append(item)
                formatted_data.append(fd)
            else:
                formatted_data.append(datalist)

        long_format = list(zip(*formatted_data))  # transform into list of rows

        predictions_table = wandb.Table(columns=table_columns, data=long_format)
        self.experiment.log({table_name: predictions_table})



import torch.nn as nn
import torch.nn.functional as F
from clipzyme.utils.registry import register_object
from clipzyme.models.abstract import AbstractModel
from torch_geometric.nn.conv.gatv2_conv import GATv2Conv
from torch_scatter import scatter
from clipzyme.utils.pyg import unbatch


@register_object("gatv2conv", "model")
class GATv2Op(AbstractModel):
    """
    https://arxiv.org/abs/2105.14491
    """

    def __init__(self, args):
        super(GATv2Op, self).__init__()
        self.args = args
        self.encoder = GATv2Conv(
            in_channels=args.num_chan,
            out_channels=args.gat_output_dim,
            heads=args.gat_num_heads,
            concat=args.gat_concat,
            negative_slope=args.gat_negative_slope,
            dropout=args.dropout,
            add_self_loops=args.gat_add_self_loops,
            edge_dim=args.gat_edge_dim,
            fill_value=args.gat_fill_value,
            bias=args.gat_bias,
            share_weights=args.gat_share_weights,
        )
        self.use_edge_features = args.gat_edge_dim is not None

    def forward(self, graph):
        output = {}
        node_features = graph.x
        edge_index = graph.edge_index
        edge_features = graph.edge_attr
        num_nodes = len(node_features)
        # default: (num_nodes, num_heads * out_chan );
        # bipartite (num_target_nodes, num_heads * out_chan )
        if self.use_edge_features:
            encoded_features = self.encoder.forward(
                node_features.float(), edge_index, edge_features.float()
            )
        else:
            encoded_features = self.encoder.forward(node_features.float(), edge_index)
        graph_features = scatter(encoded_features, graph.batch, dim=0, reduce="add")
        encoded_features = unbatch(encoded_features, graph.batch)
        output["node_features"] = encoded_features
        output["graph_features"] = graph_features
        output["hidden"] = output["graph_features"]
        return output

    @staticmethod
    def add_args(parser) -> None:
        """Add class specific args

        Args:
            parser (argparse.ArgumentParser): argument parser
        """
        parser.add_argument(
            "--gat_output_dim", type=int, nargs="*", help="Size of each output sample."
        )
        parser.add_argument(
            "--gat_num_heads",
            type=int,
            default=1,
            help="Number of multi-head-attentions.",
        )
        parser.add_argument(
            "--gat_concat",
            action="store_true",
            default=False,
            help="If set to False, the multi-head attentions are averaged instead of concatenated.",
        )
        parser.add_argument(
            "--gat_negative_slope",
            type=float,
            default=0.2,
            help="LeakyReLU angle of the negative slope.",
        )
        parser.add_argument(
            "--gat_add_self_loops",
            action="store_true",
            default=False,
            help="If set to False, will not add self-loops to the input graph.",
        )
        parser.add_argument(
            "--gat_edge_dim",
            type=int,
            default=None,
            help="Edge feature dimensionality (in case there are any).",
        )
        parser.add_argument(
            "--gat_fill_value",
            type=str,
            default="mean",
            help="The way to generate edge features of self-loops (in case edge_dim != None). If given as float or torch.Tensor, edge features of self-loops will be directly given by fill_value. If given as str, edge features of self-loops are computed by aggregating all features of edges that point to the specific node, according to a reduce operation. [add, mean, min, max, mul].",
        )
        parser.add_argument(
            "--gat_bias",
            action="store_true",
            default=False,
            help="If set to False, the layer will not learn an additive bias.",
        )
        parser.add_argument(
            "--gat_share_weights",
            action="store_true",
            default=False,
            help="If set to True, the same matrix will be applied to the source and the target node of every edge",
        )


@register_object("gatv2", "model")
class GAT(AbstractModel):
    """
    https://arxiv.org/abs/2105.14491
    """

    def __init__(self, args):
        super().__init__()

        n_hidden = args.gat_hidden_dim
        self.num_layers = args.gat_num_layers
        n_heads = args.gat_num_heads
        self.use_edge_features = args.gat_edge_dim is not None

        self.convs = nn.ModuleList()
        self.norms = nn.ModuleList()

        self.node_encoder = nn.Linear(args.gat_node_dim, n_hidden)

        for i in range(self.num_layers):
            in_hidden = n_hidden
            out_hidden = n_hidden
            # bias = i == n_layers - 1
            if (i == 0) and (self.use_edge_features):
                layer = GATv2Conv(
                    in_channels=in_hidden,
                    out_channels=out_hidden // n_heads,
                    heads=n_heads,
                    concat=True,  # args.gat_concat,
                    negative_slope=args.gat_negative_slope,
                    dropout=args.dropout,
                    add_self_loops=args.gat_add_self_loops,
                    edge_dim=args.gat_edge_dim,
                    fill_value=args.gat_fill_value,
                    bias=args.gat_bias,
                    share_weights=args.gat_share_weights,
                )
            else:
                layer = GATv2Conv(
                    in_channels=in_hidden,
                    out_channels=out_hidden // n_heads,
                    heads=n_heads,
                    concat=True,  # args.gat_concat,
                    negative_slope=args.gat_negative_slope,
                    dropout=args.dropout,
                    add_self_loops=args.gat_add_self_loops,
                    fill_value=args.gat_fill_value,
                    bias=args.gat_bias,
                    share_weights=args.gat_share_weights,
                )
            self.convs.append(layer)
            self.norms.append(nn.BatchNorm1d(out_hidden))

        self.input_drop = nn.Dropout(args.dropout)
        self.dropout = nn.Dropout(args.dropout)

        self.pool_type = args.gat_pool

    def forward(self, graph):
        output = {}

        h = graph.x.float()
        h = self.node_encoder(h)
        h = F.relu(h, inplace=True)
        h = self.input_drop(h)

        h_last = None

        for i in range(self.num_layers):
            if (i == 0) and self.use_edge_features:
                h = self.convs[i](h, graph.edge_index, graph.edge_attr.float())
            else:
                h = self.convs[i](h, graph.edge_index)

            if h_last is not None:
                h += h_last[: h.shape[0], :]

            h_last = h

            h = self.norms[i](h)
            h = F.relu(h, inplace=True)
            h = self.dropout(h)

        output["node_features"] = h
        if self.pool_type != "none":
            output["graph_features"] = scatter(
                h, graph.batch, dim=0, reduce=self.pool_type
            )
            output["hidden"] = output["graph_features"]
        return output

    @staticmethod
    def add_args(parser) -> None:
        """Add class specific args

        Args:
            parser (argparse.ArgumentParser): argument parser
        """
        parser.add_argument(
            "--gat_num_layers",
            type=int,
            default=1,
            help="Number of layers in GNN, equivalently number of convolution iterations.",
        )
        parser.add_argument(
            "--gat_hidden_dim",
            type=int,
            default=None,
            help="Dimension of hidden layers (node features)",
        )
        parser.add_argument(
            "--gat_num_heads",
            type=int,
            default=1,
            help="Number of multi-head-attentions.",
        )
        parser.add_argument(
            "--gat_concat",
            action="store_true",
            default=False,
            help="If set to False, the multi-head attentions are averaged instead of concatenated.",
        )
        parser.add_argument(
            "--gat_negative_slope",
            type=float,
            default=0.2,
            help="LeakyReLU angle of the negative slope.",
        )
        parser.add_argument(
            "--gat_add_self_loops",
            action="store_true",
            default=False,
            help="If set to False, will not add self-loops to the input graph.",
        )
        parser.add_argument(
            "--gat_node_dim",
            type=int,
            default=None,
            help="Node feature dimensionality.",
        )
        parser.add_argument(
            "--gat_edge_dim",
            type=int,
            default=None,
            help="Edge feature dimensionality (in case there are any).",
        )
        parser.add_argument(
            "--gat_fill_value",
            type=str,
            default="mean",
            help="The way to generate edge features of self-loops (in case edge_dim != None). If given as float or torch.Tensor, edge features of self-loops will be directly given by fill_value. If given as str, edge features of self-loops are computed by aggregating all features of edges that point to the specific node, according to a reduce operation. [add, mean, min, max, mul].",
        )
        parser.add_argument(
            "--gat_bias",
            action="store_true",
            default=False,
            help="If set to False, the layer will not learn an additive bias.",
        )
        parser.add_argument(
            "--gat_share_weights",
            action="store_true",
            default=False,
            help="If set to True, the same matrix will be applied to the source and the target node of every edge",
        )
        parser.add_argument(
            "--gat_pool",
            type=str,
            choices=["none", "sum", "mul", "mean", "min", "max"],
            default="sum",
            help="Type of pooling to do to obtain graph features",
        )



# -*- coding: utf-8 -*-
#
# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Convert molecules into DGLGraphs
#
# pylint: disable= no-member, arguments-differ, invalid-name
#
# Adapted from: https://github.com/awslabs/dgl-lifesci/blob/master/python/dgllife/data/uspto.py

import errno
import numpy as np
import os
import random
import torch

from collections import defaultdict
from copy import deepcopy
from functools import partial
from itertools import combinations
from multiprocessing import Pool
from rdkit import Chem
from rdkit.Chem import rdmolops, rdmolfiles
from tqdm import tqdm
from torch_geometric.data import Batch
from clipzyme.utils.pyg import from_smiles, from_mapped_smiles

from functools import partial
import torch

from sklearn.neighbors import NearestNeighbors

try:
    from molvs import Standardizer
except ImportError as e:
    print("MolVS is not installed, which is required for candidate ranking")


BOND_TYPE = {
    1: Chem.rdchem.BondType.SINGLE,
    2: Chem.rdchem.BondType.DOUBLE,
    3: Chem.rdchem.BondType.TRIPLE,
    1.5: Chem.rdchem.BondType.AROMATIC,
}

clean_rxns_postsani = [
    # two adjacent aromatic nitrogens should allow for H shift
    Chem.AllChem.ReactionFromSmarts("[n;H1;+0:1]:[n;H0;+1:2]>>[n;H0;+0:1]:[n;H0;+0:2]"),
    # two aromatic nitrogens separated by one should allow for H shift
    Chem.AllChem.ReactionFromSmarts(
        "[n;H1;+0:1]:[c:3]:[n;H0;+1:2]>>[n;H0;+0:1]:[*:3]:[n;H0;+0:2]"
    ),
    Chem.AllChem.ReactionFromSmarts("[#7;H0;+:1]-[O;H1;+0:2]>>[#7;H0;+:1]-[O;H0;-:2]"),
    # neutralize C(=O)[O-]
    Chem.AllChem.ReactionFromSmarts(
        "[C;H0;+0:1](=[O;H0;+0:2])[O;H0;-1:3]>>[C;H0;+0:1](=[O;H0;+0:2])[O;H1;+0:3]"
    ),
    # turn neutral halogens into anions EXCEPT HCl
    Chem.AllChem.ReactionFromSmarts("[I,Br,F;H1;D0;+0:1]>>[*;H0;-1:1]"),
    # inexplicable nitrogen anion in reactants gets fixed in prods
    Chem.AllChem.ReactionFromSmarts("[N;H0;-1:1]([C:2])[C:3]>>[N;H1;+0:1]([*:2])[*:3]"),
]


def robust_edit_mol(rmol, edits):
    """Simulate reaction via graph editing

    Parameters
    ----------
    rmol : rdkit.Chem.rdchem.Mol
        RDKit molecule instance for the reactants
    bond_changes : list of 3-tuples
        Each tuple is of form (atom1, atom2, change_type)
    keep_atom_map : bool
        Whether to keep atom mapping number. Default to False.

    Returns
    -------
    pred_smiles : list of str
        SMILES for the edited molecule
    """
    ################ for switching IDXes ###########################
    # edits are indices, so we need to replace the atom map number with the index
    old_index2atom_number = {}
    for atom in rmol.GetAtoms():
        old_index2atom_number[atom.GetIdx()] = int(
            atom.GetProp("molAtomMapNumber")
        )  # Update mapping dictionary

    order = sorted(
        old_index2atom_number.items(), key=lambda x: x[1]
    )  # sort by atom number because k,v = atom_idx, atom_map_number
    # old_index2new_index = {old_idx: new_idx for new_idx, (old_idx, _) in enumerate(order)}
    atom_map_number2new_index = {
        atom_map_number: new_idx for new_idx, (_, atom_map_number) in enumerate(order)
    }
    ######################################################################
    new_mol = Chem.RWMol(rmol)

    # Keep track of aromatic nitrogens, might cause explicit hydrogen issues
    aromatic_nitrogen_idx = set()
    aromatic_carbonyl_adj_to_aromatic_nH = {}
    aromatic_carbondeg3_adj_to_aromatic_nH0 = {}
    for a in new_mol.GetAtoms():
        if a.GetIsAromatic() and a.GetSymbol() == "N":
            aromatic_nitrogen_idx.add(a.GetIdx())
            for nbr in a.GetNeighbors():
                if (
                    a.GetNumExplicitHs() == 1
                    and nbr.GetSymbol() == "C"
                    and nbr.GetIsAromatic()
                    and any(b.GetBondTypeAsDouble() == 2 for b in nbr.GetBonds())
                ):
                    aromatic_carbonyl_adj_to_aromatic_nH[nbr.GetIdx()] = a.GetIdx()
                elif (
                    a.GetNumExplicitHs() == 0
                    and nbr.GetSymbol() == "C"
                    and nbr.GetIsAromatic()
                    and len(nbr.GetBonds()) == 3
                ):
                    aromatic_carbondeg3_adj_to_aromatic_nH0[nbr.GetIdx()] = a.GetIdx()
        else:
            a.SetNumExplicitHs(0)
    new_mol.UpdatePropertyCache()

    amap = {}
    for atom in rmol.GetAtoms():
        amap[atom_map_number2new_index[atom.GetIntProp("molAtomMapNumber")]] = (
            atom.GetIdx()
        )  # new index to old index

    # Apply the edits as predicted
    for x, y, t in edits:
        bond = new_mol.GetBondBetweenAtoms(amap[x], amap[y])
        a1 = new_mol.GetAtomWithIdx(amap[x])
        a2 = new_mol.GetAtomWithIdx(amap[y])
        if bond is not None:
            new_mol.RemoveBond(amap[x], amap[y])

            # Are we losing a bond on an aromatic nitrogen?
            if bond.GetBondTypeAsDouble() == 1.0:
                if amap[x] in aromatic_nitrogen_idx:
                    if a1.GetTotalNumHs() == 0:
                        a1.SetNumExplicitHs(1)
                    elif a1.GetFormalCharge() == 1:
                        a1.SetFormalCharge(0)
                elif amap[y] in aromatic_nitrogen_idx:
                    if a2.GetTotalNumHs() == 0:
                        a2.SetNumExplicitHs(1)
                    elif a2.GetFormalCharge() == 1:
                        a2.SetFormalCharge(0)

            # Are we losing a c=O bond on an aromatic ring? If so, remove H from adjacent nH if appropriate
            if bond.GetBondTypeAsDouble() == 2.0:
                if amap[x] in aromatic_carbonyl_adj_to_aromatic_nH:
                    new_mol.GetAtomWithIdx(
                        aromatic_carbonyl_adj_to_aromatic_nH[amap[x]]
                    ).SetNumExplicitHs(0)
                elif amap[y] in aromatic_carbonyl_adj_to_aromatic_nH:
                    new_mol.GetAtomWithIdx(
                        aromatic_carbonyl_adj_to_aromatic_nH[amap[y]]
                    ).SetNumExplicitHs(0)

        if t > 0:
            new_mol.AddBond(amap[x], amap[y], BOND_TYPE[t])

            # Special alkylation case?
            if t == 1:
                if amap[x] in aromatic_nitrogen_idx:
                    if a1.GetTotalNumHs() == 1:
                        a1.SetNumExplicitHs(0)
                    else:
                        a1.SetFormalCharge(1)
                elif amap[y] in aromatic_nitrogen_idx:
                    if a2.GetTotalNumHs() == 1:
                        a2.SetNumExplicitHs(0)
                    else:
                        a2.SetFormalCharge(1)

            # Are we getting a c=O bond on an aromatic ring? If so, add H to adjacent nH0 if appropriate
            if t == 2:
                if amap[x] in aromatic_carbondeg3_adj_to_aromatic_nH0:
                    new_mol.GetAtomWithIdx(
                        aromatic_carbondeg3_adj_to_aromatic_nH0[amap[x]]
                    ).SetNumExplicitHs(1)
                elif amap[y] in aromatic_carbondeg3_adj_to_aromatic_nH0:
                    new_mol.GetAtomWithIdx(
                        aromatic_carbondeg3_adj_to_aromatic_nH0[amap[y]]
                    ).SetNumExplicitHs(1)

    pred_mol = new_mol.GetMol()

    # Clear formal charges to make molecules valid
    # Note: because S and P (among others) can change valence, be more flexible
    for atom in pred_mol.GetAtoms():
        atom.ClearProp("molAtomMapNumber")
        if (
            atom.GetSymbol() == "N" and atom.GetFormalCharge() == 1
        ):  # exclude negatively-charged azide
            bond_vals = sum([bond.GetBondTypeAsDouble() for bond in atom.GetBonds()])
            if bond_vals <= 3:
                atom.SetFormalCharge(0)
        elif (
            atom.GetSymbol() == "N" and atom.GetFormalCharge() == -1
        ):  # handle negatively-charged azide addition
            bond_vals = sum([bond.GetBondTypeAsDouble() for bond in atom.GetBonds()])
            if bond_vals == 3 and any(
                [nbr.GetSymbol() == "N" for nbr in atom.GetNeighbors()]
            ):
                atom.SetFormalCharge(0)
        elif atom.GetSymbol() == "N":
            bond_vals = sum([bond.GetBondTypeAsDouble() for bond in atom.GetBonds()])
            if (
                bond_vals == 4 and not atom.GetIsAromatic()
            ):  # and atom.IsInRingSize(5)):
                atom.SetFormalCharge(1)
        elif atom.GetSymbol() == "C" and atom.GetFormalCharge() != 0:
            atom.SetFormalCharge(0)
        elif atom.GetSymbol() == "O" and atom.GetFormalCharge() != 0:
            bond_vals = (
                sum([bond.GetBondTypeAsDouble() for bond in atom.GetBonds()])
                + atom.GetNumExplicitHs()
            )
            if bond_vals == 2:
                atom.SetFormalCharge(0)
        elif atom.GetSymbol() in ["Cl", "Br", "I", "F"] and atom.GetFormalCharge() != 0:
            bond_vals = sum([bond.GetBondTypeAsDouble() for bond in atom.GetBonds()])
            if bond_vals == 1:
                atom.SetFormalCharge(0)
        elif atom.GetSymbol() == "S" and atom.GetFormalCharge() != 0:
            bond_vals = sum([bond.GetBondTypeAsDouble() for bond in atom.GetBonds()])
            if bond_vals in [2, 4, 6]:
                atom.SetFormalCharge(0)
        elif (
            atom.GetSymbol() == "P"
        ):  # quartenary phosphorous should be pos. charge with 0 H
            bond_vals = [bond.GetBondTypeAsDouble() for bond in atom.GetBonds()]
            if sum(bond_vals) == 4 and len(bond_vals) == 4:
                atom.SetFormalCharge(1)
                atom.SetNumExplicitHs(0)
            elif sum(bond_vals) == 3 and len(bond_vals) == 3:  # make sure neutral
                atom.SetFormalCharge(0)
        elif atom.GetSymbol() == "B":  # quartenary boron should be neg. charge with 0 H
            bond_vals = [bond.GetBondTypeAsDouble() for bond in atom.GetBonds()]
            if sum(bond_vals) == 4 and len(bond_vals) == 4:
                atom.SetFormalCharge(-1)
                atom.SetNumExplicitHs(0)
        elif atom.GetSymbol() in ["Mg", "Zn"]:
            bond_vals = [bond.GetBondTypeAsDouble() for bond in atom.GetBonds()]
            if sum(bond_vals) == 1 and len(bond_vals) == 1:
                atom.SetFormalCharge(1)
        elif atom.GetSymbol() == "Si":
            bond_vals = [bond.GetBondTypeAsDouble() for bond in atom.GetBonds()]
            if sum(bond_vals) == len(bond_vals):
                atom.SetNumExplicitHs(max(0, 4 - len(bond_vals)))

    # Bounce to/from SMILES to try to sanitize
    pred_smiles = Chem.MolToSmiles(pred_mol)
    pred_list = pred_smiles.split(".")
    pred_mols = [Chem.MolFromSmiles(pred_smiles) for pred_smiles in pred_list]

    for i, mol in enumerate(pred_mols):
        # Check if we failed/succeeded in previous step
        if mol is None:
            # print('##### Unparseable mol: {}'.format(pred_list[i]))
            continue

        # Else, try post-sanitiztion fixes in structure
        mol = Chem.MolFromSmiles(Chem.MolToSmiles(mol))
        if mol is None:
            continue
        for rxn in clean_rxns_postsani:
            out = rxn.RunReactants((mol,))
            if out:
                try:
                    Chem.SanitizeMol(out[0][0])
                    pred_mols[i] = Chem.MolFromSmiles(Chem.MolToSmiles(out[0][0]))
                except Exception as e:
                    pass
                    # print(e)
                    # print('Could not sanitize postsani reaction product: {}'.format(Chem.MolToSmiles(out[0][0])))
                    # print('Original molecule was: {}'.format(Chem.MolToSmiles(mol)))
    pred_smiles = [
        Chem.MolToSmiles(pred_mol) for pred_mol in pred_mols if pred_mol is not None
    ]

    return pred_smiles


def get_bond_changes(reaction):
    """Get the bond changes in a reaction.

    Parameters
    ----------
    reaction : str
        SMILES for a reaction, e.g. [CH3:14][NH2:15].[N+:1](=[O:2])([O-:3])[c:4]1[cH:5][c:6]([C:7]
        (=[O:8])[OH:9])[cH:10][cH:11][c:12]1[Cl:13].[OH2:16]>>[N+:1](=[O:2])([O-:3])[c:4]1[cH:5]
        [c:6]([C:7](=[O:8])[OH:9])[cH:10][cH:11][c:12]1[NH:15][CH3:14]. It consists of reactants,
        products and the atom mapping.

    Returns
    -------
    bond_changes : set of 3-tuples
        Each tuple consists of (atom1, atom2, change type)
        There are 5 possible values for change type. 0 for losing the bond, and 1, 2, 3, 1.5
        separately for forming a single, double, triple or aromatic bond.
    """
    reactants = Chem.MolFromSmiles(reaction.split(">")[0])
    products = Chem.MolFromSmiles(reaction.split(">")[2])

    conserved_maps = [
        a.GetProp("molAtomMapNumber")
        for a in products.GetAtoms()
        if a.HasProp("molAtomMapNumber")
    ]
    bond_changes = set()  # keep track of bond changes

    # Look at changed bonds
    bonds_prev = {}
    for bond in reactants.GetBonds():
        nums = sorted(
            [
                bond.GetBeginAtom().GetProp("molAtomMapNumber"),
                bond.GetEndAtom().GetProp("molAtomMapNumber"),
            ]
        )
        if (nums[0] not in conserved_maps) and (nums[1] not in conserved_maps):
            continue
        bonds_prev["{}~{}".format(nums[0], nums[1])] = bond.GetBondTypeAsDouble()
    bonds_new = {}
    for bond in products.GetBonds():
        nums = sorted(
            [
                bond.GetBeginAtom().GetProp("molAtomMapNumber"),
                bond.GetEndAtom().GetProp("molAtomMapNumber"),
            ]
        )
        bonds_new["{}~{}".format(nums[0], nums[1])] = bond.GetBondTypeAsDouble()

    for bond in bonds_prev:
        if bond not in bonds_new:
            # lost bond
            bond_changes.add((bond.split("~")[0], bond.split("~")[1], 0.0))
        else:
            if bonds_prev[bond] != bonds_new[bond]:
                # changed bond
                bond_changes.add(
                    (bond.split("~")[0], bond.split("~")[1], bonds_new[bond])
                )
    for bond in bonds_new:
        if bond not in bonds_prev:
            # new bond
            bond_changes.add((bond.split("~")[0], bond.split("~")[1], bonds_new[bond]))

    return bond_changes


def edit_mol(reactant_mols, edits, product_info, mode="train", return_full_mol=False):
    """Simulate reaction via graph editing

    Parameters
    ----------
    reactant_mols : rdkit.Chem.rdchem.Mol
        RDKit molecule instances for reactants.
    edits : list of 4-tuples
        Bond changes for getting the product out of the reactants in a reaction.
        Each 4-tuple is of form (atom1, atom2, change_type, score), where atom1
        and atom2 are the end atoms to form or lose a bond, change_type is the
        type of bond change and score represents the confidence for the bond change
        by a model.
    product_info : dict
        product_info['atoms'] gives a set of atom ids in the ground truth product molecule.
    return_full_mol : bool
        keep atom-mapping and do not remove fragments not in product (use for generating products to rank)

    Returns
    -------
    str
        SMILES for the main products
    """
    ################ for switching IDXes ###########################
    old_index2atom_number = {}
    for atom in reactant_mols.GetAtoms():
        old_index2atom_number[atom.GetIdx()] = int(
            atom.GetProp("molAtomMapNumber")
        )  # Update mapping dictionary

    order = sorted(
        old_index2atom_number.items(), key=lambda x: x[1]
    )  # sort by atom number because k,v = atom_idx, atom_map_number
    old_index2new_index = {
        old_idx: new_idx for new_idx, (old_idx, _) in enumerate(order)
    }
    new_index2old_index = {
        new_idx: old_idx for new_idx, (old_idx, _) in enumerate(order)
    }
    atom_map_number2new_index = {
        atom_map_number: new_idx for new_idx, (_, atom_map_number) in enumerate(order)
    }
    ######################################################################

    bond_change_to_type = {
        1: Chem.rdchem.BondType.SINGLE,
        2: Chem.rdchem.BondType.DOUBLE,
        3: Chem.rdchem.BondType.TRIPLE,
        1.5: Chem.rdchem.BondType.AROMATIC,
    }

    new_mol = Chem.RWMol(reactant_mols)
    [atom.SetNumExplicitHs(0) for atom in new_mol.GetAtoms()]

    for atom1, atom2, change_type, score in edits:
        atom1, atom2 = new_index2old_index[atom1], new_index2old_index[atom2]
        bond = new_mol.GetBondBetweenAtoms(atom1, atom2)
        if bond is not None:
            new_mol.RemoveBond(atom1, atom2)
        if change_type > 0:
            new_mol.AddBond(atom1, atom2, bond_change_to_type[change_type])

    pred_mol = new_mol.GetMol()
    pred_smiles = Chem.MolToSmiles(pred_mol)

    if return_full_mol:
        return pred_smiles

    pred_list = pred_smiles.split(".")
    pred_mols = []
    for pred_smiles in pred_list:
        mol = Chem.MolFromSmiles(pred_smiles)
        if mol is None:
            continue
        atom_set = set(
            [atom_map_number2new_index[atom.GetAtomMapNum()] for atom in mol.GetAtoms()]
        )

        if mode == "train":
            if (
                len(atom_set & product_info["atoms"]) == 0
            ):  # no overlap with gold product atoms
                continue
        for atom in mol.GetAtoms():
            atom.SetAtomMapNum(0)
        pred_mols.append(mol)

    return ".".join(sorted([Chem.MolToSmiles(mol) for mol in pred_mols]))


def get_product_smiles(
    reactant_mols, edits, product_info, mode="train", return_full_mol=False
):
    """Get the product smiles of the reaction

    Parameters
    ----------
    reactant_mols : rdkit.Chem.rdchem.Mol
        RDKit molecule instances for reactants.
    edits : list of 4-tuples
        Bond changes for getting the product out of the reactants in a reaction.
        Each 4-tuple is of form (atom1, atom2, change_type, score), where atom1
        and atom2 are the end atoms to form or lose a bond, change_type is the
        type of bond change and score represents the confidence for the bond change
        by a model.
    product_info : dict
        product_info['atoms'] gives a set of atom ids in the ground truth product molecule.

    Returns
    -------
    str
        SMILES for the main products
    """
    smiles = edit_mol(reactant_mols, edits, product_info, mode, return_full_mol)
    if len(smiles) != 0:
        return smiles
    try:
        Chem.Kekulize(reactant_mols)
    except Exception as e:
        return smiles
    return edit_mol(reactant_mols, edits, product_info, mode, return_full_mol)


def is_connected_change_combo(combo_ids, cand_change_adj):
    """Check whether the combo of bond changes yields a connected component.

    Parameters
    ----------
    combo_ids : tuple of int
        Ids for bond changes in the combination.
    cand_change_adj : bool ndarray of shape (N, N)
        Adjacency matrix for candidate bond changes. Two candidate bond
        changes are considered adjacent if they share a common atom.
        * N for the number of candidate bond changes.

    Returns
    -------
    bool
        Whether the combo of bond changes yields a connected component
    """
    if len(combo_ids) == 1:
        return True
    multi_hop_adj = np.linalg.matrix_power(
        cand_change_adj[combo_ids, :][:, combo_ids], len(combo_ids) - 1
    )
    # The combo is connected if the distance between
    # any pair of bond changes is within len(combo) - 1

    return np.all(multi_hop_adj)


def is_valid_combo(combo_changes, reactant_info):
    """Whether the combo of bond changes is chemically valid.

    Parameters
    ----------
    combo_changes : list of 4-tuples
        Each tuple consists of atom1, atom2, type of bond change (in the form of related
        valence) and score for the change.
    reactant_info : dict
        Reaction-related information of reactants

    Returns
    -------
    bool
        Whether the combo of bond changes is chemically valid.
    """
    num_atoms = len(reactant_info["free_val"])
    force_even_parity = np.zeros((num_atoms,), dtype=bool)
    force_odd_parity = np.zeros((num_atoms,), dtype=bool)
    pair_seen = defaultdict(bool)
    free_val_tmp = reactant_info["free_val"].copy()
    for atom1, atom2, change_type, score in combo_changes:
        if pair_seen[(atom1, atom2)]:
            # A pair of atoms cannot have two types of changes. Even if we
            # randomly pick one, that will be reduced to a combo of less changes
            return False
        pair_seen[(atom1, atom2)] = True

        # Special valence rules
        atom1_type_val = atom2_type_val = change_type
        if change_type == 2:
            # to form a double bond
            if reactant_info["is_o"][atom1]:
                if reactant_info["is_c2_of_pyridine"][atom2]:
                    atom2_type_val = 1.0
                elif reactant_info["is_p"][atom2]:
                    # don't count information of =o toward valence
                    # but require odd valence parity
                    atom2_type_val = 0.0
                    force_odd_parity[atom2] = True
                elif reactant_info["is_s"][atom2]:
                    atom2_type_val = 0.0
                    force_even_parity[atom2] = True
            elif reactant_info["is_o"][atom2]:
                if reactant_info["is_c2_of_pyridine"][atom1]:
                    atom1_type_val = 1.0
                elif reactant_info["is_p"][atom1]:
                    atom1_type_val = 0.0
                    force_odd_parity[atom1] = True
                elif reactant_info["is_s"][atom1]:
                    atom1_type_val = 0.0
                    force_even_parity[atom1] = True
            elif reactant_info["is_n"][atom1] and reactant_info["is_p"][atom2]:
                atom2_type_val = 0.0
                force_odd_parity[atom2] = True
            elif reactant_info["is_n"][atom2] and reactant_info["is_p"][atom1]:
                atom1_type_val = 0.0
                force_odd_parity[atom1] = True
            elif reactant_info["is_p"][atom1] and reactant_info["is_c"][atom2]:
                atom1_type_val = 0.0
                force_odd_parity[atom1] = True
            elif reactant_info["is_p"][atom2] and reactant_info["is_c"][atom1]:
                atom2_type_val = 0.0
                force_odd_parity[atom2] = True

        reactant_pair_val = reactant_info["pair_to_bond_val"].get((atom1, atom2), None)
        if reactant_pair_val is not None:
            free_val_tmp[atom1] += reactant_pair_val - atom1_type_val
            free_val_tmp[atom2] += reactant_pair_val - atom2_type_val
        else:
            free_val_tmp[atom1] -= atom1_type_val
            free_val_tmp[atom2] -= atom2_type_val

    free_val_tmp = np.array(
        [l.item() if isinstance(l, torch.Tensor) else l for l in free_val_tmp]
    )

    # False if 1) too many connections 2) sulfur valence not even
    # 3) phosphorous valence not odd
    if (
        any(free_val_tmp < 0)
        or any(aval % 2 != 0 for aval in free_val_tmp[force_even_parity])
        or any(aval % 2 != 1 for aval in free_val_tmp[force_odd_parity])
    ):
        return False
    return True


def bookkeep_reactant(mol):  # , candidate_pairs):
    """Bookkeep reaction-related information of reactants.

    Parameters
    ----------
    mol : rdkit.Chem.rdchem.Mol
        RDKit molecule instance for reactants.
    candidate_pairs : list of 2-tuples
        Pairs of atoms that ranked high by a model for reaction center prediction.
        By assumption, the two atoms are different and the first atom has a smaller
        index than the second.

    Returns
    -------
    info : dict
        Reaction-related information of reactants
    """
    ################ for switching IDXes ###########################
    # edits are indices, so we need to replace the atom map number with the index
    old_index2atom_number = {}
    for atom in mol.GetAtoms():
        old_index2atom_number[atom.GetIdx()] = int(
            atom.GetProp("molAtomMapNumber")
        )  # Update mapping dictionary

    order = sorted(
        old_index2atom_number.items(), key=lambda x: x[1]
    )  # sort by atom number because k,v = atom_idx, atom_map_number
    old_index2new_index = {
        old_idx: new_idx for new_idx, (old_idx, _) in enumerate(order)
    }
    # atom_map_number2new_index = {atom_map_number: new_idx for new_idx, (_, atom_map_number) in enumerate(order)}
    ######################################################################

    num_atoms = mol.GetNumAtoms()
    info = {
        # free valence of atoms
        "free_val": [0 for _ in range(num_atoms)],
        # Whether it is a carbon atom
        "is_c": [False for _ in range(num_atoms)],
        # Whether it is a carbon atom connected to a nitrogen atom in pyridine
        "is_c2_of_pyridine": [False for _ in range(num_atoms)],
        # Whether it is a phosphorous atom
        "is_p": [False for _ in range(num_atoms)],
        # Whether it is a sulfur atom
        "is_s": [False for _ in range(num_atoms)],
        # Whether it is an oxygen atom
        "is_o": [False for _ in range(num_atoms)],
        # Whether it is a nitrogen atom
        "is_n": [False for _ in range(num_atoms)],
        "pair_to_bond_val": dict(),
        "ring_bonds": set(),
    }

    # bookkeep atoms
    for j, atom in enumerate(mol.GetAtoms()):
        j = old_index2new_index[j]
        info["free_val"][j] += atom.GetTotalNumHs() + abs(atom.GetFormalCharge())
        # An aromatic carbon atom next to an aromatic nitrogen atom can get a
        # carbonyl b/c of bookkeeping of hydroxypyridines
        if atom.GetSymbol() == "C":
            info["is_c"][j] = True
            if atom.GetIsAromatic():
                for nbr in atom.GetNeighbors():
                    if nbr.GetSymbol() == "N" and nbr.GetDegree() == 2:
                        info["is_c2_of_pyridine"][j] = True
                        break
        # A nitrogen atom should be allowed to become positively charged
        elif atom.GetSymbol() == "N":
            info["free_val"][j] += 1 - atom.GetFormalCharge()
            info["is_n"][j] = True
        # Phosphorous atoms can form a phosphonium
        elif atom.GetSymbol() == "P":
            info["free_val"][j] += 1 - atom.GetFormalCharge()
            info["is_p"][j] = True
        elif atom.GetSymbol() == "O":
            info["is_o"][j] = True
        elif atom.GetSymbol() == "S":
            info["is_s"][j] = True

    # bookkeep bonds
    for bond in mol.GetBonds():
        atom1, atom2 = bond.GetBeginAtomIdx(), bond.GetEndAtomIdx()

        atom1, atom2 = old_index2new_index[atom1], old_index2new_index[atom2]
        atom1, atom2 = min(atom1, atom2), max(atom1, atom2)
        type_val = bond.GetBondTypeAsDouble()
        info["pair_to_bond_val"][(atom1, atom2)] = type_val
        # if (atom1, atom2) in candidate_pairs:
        #     info['free_val'][atom1] += type_val
        #     info['free_val'][atom2] += type_val
        if bond.IsInRing():
            info["ring_bonds"].add((atom1, atom2))

    return info


def bookkeep_product(mol):
    """Bookkeep reaction-related information of atoms/bonds in products

    Parameters
    ----------
    mol : rdkit.Chem.rdchem.Mol
        RDKit molecule instance for products.

    Returns
    -------
    info : dict
        Reaction-related information of atoms/bonds in products
    """
    ################ for switching IDXes ###########################
    old_index2atom_number = {}
    for atom in mol.GetAtoms():
        old_index2atom_number[atom.GetIdx()] = int(
            atom.GetProp("molAtomMapNumber")
        )  # Update mapping dictionary

    order = sorted(
        old_index2atom_number.items(), key=lambda x: x[1]
    )  # sort by atom number because k,v = atom_idx, atom_map_number
    old_index2new_index = {
        old_idx: new_idx for new_idx, (old_idx, _) in enumerate(order)
    }
    atom_map_number2new_index = {
        atom_map_number: new_idx for new_idx, (_, atom_map_number) in enumerate(order)
    }
    ######################################################################
    info = {"atoms": set()}
    for atom in mol.GetAtoms():
        info["atoms"].add(atom_map_number2new_index[atom.GetAtomMapNum()])

    return info


def pre_process_one_reaction(
    info, num_candidate_bond_changes, max_num_bond_changes, max_num_change_combos, mode
):
    """Pre-process one reaction for candidate ranking.

    Parameters
    ----------
    info : 4-tuple
        * candidate_bond_changes : list of tuples
            The candidate bond changes for the reaction
        * real_bond_changes : list of tuples
            The real bond changes for the reaction
        * reactant_mol : rdkit.Chem.rdchem.Mol
            RDKit molecule instance for reactants
        * product_mol : rdkit.Chem.rdchem.Mol
            RDKit molecule instance for product
    num_candidate_bond_changes : int
        Number of candidate bond changes to consider for the ground truth reaction.
    max_num_bond_changes : int
        Maximum number of bond changes per reaction.
    max_num_change_combos : int
        Number of bond change combos to consider for each reaction.
    mode : str
        Whether the dataset is to be used for training, validation or test.

    Returns
    -------
    valid_candidate_combos : list
        valid_candidate_combos[i] gives a list of tuples, which is the i-th valid combo
        of candidate bond changes for the reaction.
    candidate_bond_changes : list of 4-tuples
        Refined candidate bond changes considered for combos.
    reactant_info : dict
        Reaction-related information of reactants.
    """
    assert mode in [
        "train",
        "val",
        "test",
    ], "Expect mode to be 'train' or 'val' or 'test', got {}".format(mode)
    candidate_bond_changes_, real_bond_changes, reactant_mol, product_mol = info
    # candidate_pairs = [(atom1, atom2) for (atom1, atom2, _, _) in candidate_bond_changes_]
    reactant_info = bookkeep_reactant(reactant_mol)  # , candidate_pairs)
    if mode == "train":
        product_info = bookkeep_product(product_mol)

    # Filter out candidate new bonds already in reactants
    candidate_bond_changes = []
    count = 0
    for atom1, atom2, change_type, score in candidate_bond_changes_:
        if (
            (atom1, atom2) not in reactant_info["pair_to_bond_val"] and change_type > 0
        ) or (
            (atom1, atom2) in reactant_info["pair_to_bond_val"]
            and reactant_info["pair_to_bond_val"][(atom1, atom2)] != change_type
        ):
            candidate_bond_changes.append((atom1, atom2, change_type, score))
            count += 1
            if count == num_candidate_bond_changes:
                break

    # Check if two bond changes have atom in common
    cand_change_adj = np.eye(len(candidate_bond_changes), dtype=bool)
    for i in range(len(candidate_bond_changes)):
        atom1_1, atom1_2, _, _ = candidate_bond_changes[i]
        for j in range(i + 1, len(candidate_bond_changes)):
            atom2_1, atom2_2, _, _ = candidate_bond_changes[j]
            if (
                atom1_1 == atom2_1
                or atom1_1 == atom2_2
                or atom1_2 == atom2_1
                or atom1_2 == atom2_2
            ):
                cand_change_adj[i, j] = cand_change_adj[j, i] = True

    # Enumerate combinations of k candidate bond changes and record
    # those that are connected and chemically valid
    valid_candidate_combos = []
    cand_change_ids = range(len(candidate_bond_changes))
    for k in range(1, max_num_bond_changes + 1):
        for combo_ids in combinations(cand_change_ids, k):
            # Check if the changed bonds form a connected component
            if not is_connected_change_combo(combo_ids, cand_change_adj):
                continue
            combo_changes = [candidate_bond_changes[j] for j in combo_ids]
            # Check if the combo is chemically valid
            if is_valid_combo(combo_changes, reactant_info):
                valid_candidate_combos.append(combo_changes)

    if mode == "train":
        random.shuffle(valid_candidate_combos)
        # Index for the combo of candidate bond changes
        # that is equivalent to the gold combo
        real_combo_id = -1
        for j, combo_changes in enumerate(valid_candidate_combos):
            if set(
                [
                    (atom1, atom2, change_type)
                    for (atom1, atom2, change_type, score) in combo_changes
                ]
            ) == set(real_bond_changes):
                real_combo_id = j
                break

        # If we fail to find the real combo, make it the first entry
        if real_combo_id == -1:
            valid_candidate_combos = [
                [
                    (atom1, atom2, change_type, 0.0)
                    for (atom1, atom2, change_type) in real_bond_changes
                ]
            ] + valid_candidate_combos
        else:
            valid_candidate_combos[0], valid_candidate_combos[real_combo_id] = (
                valid_candidate_combos[real_combo_id],
                valid_candidate_combos[0],
            )

        product_smiles = get_product_smiles(
            reactant_mol, valid_candidate_combos[0], product_info, mode
        )
        if len(product_smiles) > 0:
            # Remove combos yielding duplicate products
            product_smiles = set([product_smiles])
            new_candidate_combos = [valid_candidate_combos[0]]

            count = 0
            for combo in valid_candidate_combos[1:]:
                smiles = get_product_smiles(reactant_mol, combo, product_info, mode)
                if smiles in product_smiles or len(smiles) == 0:
                    continue
                product_smiles.add(smiles)
                new_candidate_combos.append(combo)
                count += 1
                if count == max_num_change_combos:
                    break
            valid_candidate_combos = new_candidate_combos
    valid_candidate_combos = valid_candidate_combos[:max_num_change_combos]

    candidate_smiles = [
        get_product_smiles(reactant_mol, combo, None, mode="test", return_full_mol=True)
        for combo in valid_candidate_combos
    ]

    return (
        valid_candidate_combos,
        candidate_bond_changes,
        reactant_info,
        candidate_smiles,
    )


def get_batch_candidate_bonds(reaction_strings, preds, batch_ids):
    all_candidates = []
    batch_sizes = torch.bincount(batch_ids)  # Compute the number of nodes in each graph
    node_offset = 0  # Keep track of the starting node index for each graph

    for idx, reaction in enumerate(reaction_strings):
        num_nodes = batch_sizes[idx]
        graph_preds = preds[
            node_offset : node_offset + num_nodes, node_offset : node_offset + num_nodes
        ]
        candidates = tensor_to_tuples(graph_preds)
        all_candidates.append(candidates)
        node_offset += num_nodes  # Move the offset to the start of the next graph

    return all_candidates


def tensor_to_tuples(t):
    N = t.shape[0]
    K = t.shape[2]

    # Get indices of the original tensor
    indices = torch.triu_indices(N, N, offset=1).T  # upper triangular indices
    maxes, arg_maxes = torch.max(t, dim=2)  # (N, N)

    # Convert to a list of tuples
    tuples = [
        (int(row), int(col), arg_maxes[row, col].item(), maxes[row, col].item())
        for row, col in indices
    ]
    tuples = sorted(tuples, key=lambda x: x[-1], reverse=True)
    return tuples


def get_atom_pair_to_bond(mol):
    """Bookkeep bonds in the reactant.

    Parameters
    ----------
    mol : rdkit.Chem.rdchem.Mol
        RDKit molecule instance for reactants.

    Returns
    -------
    pair_to_bond_type : dict
        Mapping 2-tuples of atoms to bond type. 1, 2, 3, 1.5 are
        separately for single, double, triple and aromatic bond.
    """

    ################ for switching IDXes ###########################
    old_index2atom_number = {}
    for atom in mol.GetAtoms():
        old_index2atom_number[atom.GetIdx()] = int(
            atom.GetProp("molAtomMapNumber")
        )  # Update mapping dictionary

    order = sorted(
        old_index2atom_number.items(), key=lambda x: x[1]
    )  # sort by atom number because k,v = atom_idx, atom_map_number
    old_index2new_index = {
        old_idx: new_idx for new_idx, (old_idx, _) in enumerate(order)
    }
    atom_map_number2new_index = {
        atom_map_number: new_idx for new_idx, (_, atom_map_number) in enumerate(order)
    }
    ######################################################################

    pair_to_bond_type = dict()
    for bond in mol.GetBonds():
        atom1, atom2 = bond.GetBeginAtomIdx(), bond.GetEndAtomIdx()
        atom1, atom2 = old_index2new_index[atom1], old_index2new_index[atom2]
        atom1, atom2 = min(atom1, atom2), max(atom1, atom2)
        type_val = bond.GetBondTypeAsDouble()
        pair_to_bond_type[(atom1, atom2)] = type_val

    return pair_to_bond_type


def generate_candidates_from_scores(model_output, batch, args, mode="train"):
    """Generate candidate products from predicted changes

    Parameters
    ----------
    model_ouput : dict
        Predictions from reactivity center model.
    args : Namespace
        global arguments.
    batch : dict
        Dataset mini batch of reaction samples.
    mode : str
        Whether ground truth products are known (training) or not (inference)

    Returns
    -------
    list_of_data_batches:
        list of DataBatch of candidate products for each reaction in the batch
    valid_candidate_combos:
        list of valid candidate combinations for each reaction in the batch
    """
    num_candidate_bond_changes = args.num_candidate_bond_changes  # core size 20
    max_num_bond_changes = args.max_num_bond_changes  # combinations 5
    max_num_change_combos_per_reaction = (
        args.max_num_change_combos_per_reaction
    )  # cutoff 500

    # returns a list of list of tuples
    candidate_bond_changes = get_batch_candidate_bonds(
        batch["reaction"], model_output["s_uv"], batch["reactants"].batch
    )
    # make bonds that are "4" -> "1.5"
    for i in range(len(candidate_bond_changes)):
        candidate_bond_changes[i] = [
            (elem[0], elem[1], 1.5, elem[3]) if elem[2] == 4 else elem
            for elem in candidate_bond_changes[i]
        ]

    # Get valid candidate products, candidate bond changes considered and reactant info
    candidate_smiles = []
    valid_candidate_combos = []
    # valid_candidate_combos, candidate_bond_changes_many, reactant_info,  = [], [], []
    for i in range(len(candidate_bond_changes)):
        real_bond_changes = [
            tuple(elem) for elem in batch["reactants"]["bond_changes"][i]
        ]
        reactant_mol = Chem.MolFromSmiles(batch["reactants"]["smiles"][i])
        if mode == "train":
            product_mol = Chem.MolFromSmiles(batch["products"]["smiles"][i])
        else:
            product_mol = None

        info = (candidate_bond_changes[i], real_bond_changes, reactant_mol, product_mol)
        (
            valid_candidate_combos_one,
            candidate_bond_changes_one,
            reactant_info_one,
            candidate_smiles_one,
        ) = pre_process_one_reaction(
            info,
            num_candidate_bond_changes,
            max_num_bond_changes,
            max_num_change_combos_per_reaction,
            mode,
        )
        valid_candidate_combos.append(valid_candidate_combos_one)
        # candidate_bond_changes_many.append(candidate_bond_changes_one)
        # reactant_info.append(reactant_info_one)
        # real_bond_changes_fake_scores = [(elem[0], elem[1], elem[2], 1000) for elem in real_bond_changes]
        candidate_smiles.append(candidate_smiles_one)

    list_of_data_batches = []
    for i, list_of_candidates in enumerate(candidate_smiles):
        data_batch = []
        for j, candidate in enumerate(list_of_candidates):
            data, _ = from_mapped_smiles(
                candidate, encode_no_edge=False, sanitize=False
            )
            # fail to convert to rdkit mol
            if (data is None) or (len(data.x) == 0):
                continue
            data.candidate_bond_change = valid_candidate_combos[i][
                j
            ]  # add the bond change tuple to the Data object
            data_batch.append(data)
        if len(data_batch) == 0:
            continue
        data_batch = Batch.from_data_list(data_batch)
        data_batch.index_in_batch = i
        list_of_data_batches.append(data_batch)

    return list_of_data_batches


def sanitize_smiles_molvs(smiles, largest_fragment=False):
    """Sanitize a SMILES with MolVS

    Parameters
    ----------
    smiles : str
        SMILES string for a molecule.
    largest_fragment : bool
        Whether to select only the largest covalent unit in a molecule with
        multiple fragments. Default to False.

    Returns
    -------
    str
        SMILES string for the sanitized molecule.
    """
    standardizer = Standardizer()
    standardizer.prefer_organic = True

    mol = Chem.MolFromSmiles(smiles)
    if mol is None:
        return smiles
    try:
        mol = standardizer.standardize(mol)  # standardize functional group reps
        if largest_fragment:
            mol = standardizer.largest_fragment(
                mol
            )  # remove product counterions/salts/etc.
        mol = standardizer.uncharge(mol)  # neutralize, e.g., carboxylic acids
    except Exception:
        pass
    return Chem.MolToSmiles(mol)


def examine_topk_candidate_product(
    topks,
    topk_combos,
    reactant_mol,
    real_bond_changes,
    product_mol,
    remove_stereochemistry=False,
):
    """Perform topk evaluation for predicting the product of a reaction

    Parameters
    ----------
    topks : list of int
        Options for top-k evaluation, e.g. [1, 3, ...].
    topk_combos : list of list
        topk_combos[i] gives the combo of valid bond changes ranked i-th,
        which is a list of 3-tuples. Each tuple is of form
        (atom1, atom2, change_type). atom1, atom2 are the atom mapping numbers - 1 of the two
        end atoms. The change_type can be 0, 1, 2, 3, 1.5, separately for losing a bond or
        forming a single, double, triple, aromatic bond.
    reactant_mol : rdkit.Chem.rdchem.Mol
        RDKit molecule instance for the reactants.
    real_bond_changes : list of tuples
        Ground truth bond changes in a reaction. Each tuple is of form (atom1, atom2,
        change_type). atom1, atom2 are the atom mapping numbers - 1 of the two
        end atoms. change_type can be 0, 1, 2, 3, 1.5, separately for losing a bond, forming
        a single, double, triple, and aromatic bond.
    product_mol : rdkit.Chem.rdchem.Mol
        RDKit molecule instance for the product.
    get_smiles : bool
        Whether to get the SMILES of candidate products.

    Returns
    -------
    found_info : dict
        Binary values indicating whether we can recover the product from the ground truth
        graph edits or top-k predicted edits
    """
    if remove_stereochemistry:
        Chem.RemoveStereochemistry(reactant_mol)
        Chem.RemoveStereochemistry(product_mol)

    found_info = defaultdict(bool)

    # Avoid corrupting the RDKit molecule instances in the dataset
    reactant_mol = deepcopy(reactant_mol)
    product_mol = deepcopy(product_mol)

    for atom in product_mol.GetAtoms():
        atom.ClearProp("molAtomMapNumber")
    product_smiles = Chem.MolToSmiles(product_mol)
    product_smiles_sanitized = set(
        sanitize_smiles_molvs(product_smiles, True).split(".")
    )
    product_smiles = set(product_smiles.split("."))

    ########### Use *true* edits to try to recover product
    # Generate product by modifying reactants with graph edits
    pred_smiles = robust_edit_mol(reactant_mol, real_bond_changes)
    pred_smiles_sanitized = set(sanitize_smiles_molvs(smiles) for smiles in pred_smiles)
    pred_smiles = set(pred_smiles)

    if not product_smiles <= pred_smiles:
        # Try again with kekulized form
        Chem.Kekulize(reactant_mol)
        pred_smiles_kek = robust_edit_mol(reactant_mol, real_bond_changes)
        pred_smiles_kek = set(pred_smiles_kek)
        if not product_smiles <= pred_smiles_kek:
            if product_smiles_sanitized <= pred_smiles_sanitized:
                # print('\nwarn: mismatch, but only due to standardization')
                found_info["ground_sanitized"] = True
            else:
                pass
                # print('\nwarn: could not regenerate product {}'.format(product_smiles))
                # print('sani product: {}'.format(product_smiles_sanitized))
                # print(Chem.MolToSmiles(reactant_mol))
                # print(Chem.MolToSmiles(product_mol))
                # print(real_bond_changes)
                # print('pred_smiles: {}'.format(pred_smiles))
                # print('pred_smiles_kek: {}'.format(pred_smiles_kek))
                # print('pred_smiles_sani: {}'.format(pred_smiles_sanitized))
        else:
            found_info["ground"] = True
            found_info["ground_sanitized"] = True
    else:
        found_info["ground"] = True
        found_info["ground_sanitized"] = True

    ########### Now use candidate edits to try to recover product
    max_topk = max(topks)
    current_rank = 0
    correct_rank = max_topk + 1
    sanitized_correct_rank = max_topk + 1
    candidate_smiles_list = []
    candidate_smiles_sanitized_list = []

    for i, combo in enumerate(topk_combos):
        prev_len_candidate_smiles = len(set(candidate_smiles_list))

        # Generate products by modifying reactants with predicted edits.
        candidate_smiles = robust_edit_mol(reactant_mol, combo)
        candidate_smiles = set(candidate_smiles)
        candidate_smiles_sanitized = set(
            sanitize_smiles_molvs(smiles) for smiles in candidate_smiles
        )

        if product_smiles_sanitized <= candidate_smiles_sanitized:
            sanitized_correct_rank = min(sanitized_correct_rank, current_rank + 1)
        if product_smiles <= candidate_smiles:
            correct_rank = min(correct_rank, current_rank + 1)

        # Record unkekulized form
        candidate_smiles_list.append(".".join(candidate_smiles))
        candidate_smiles_sanitized_list.append(".".join(candidate_smiles_sanitized))

        # Edit molecules with reactants kekulized. Sometimes previous editing fails due to
        # RDKit sanitization error (edited molecule cannot be kekulized)
        try:
            Chem.Kekulize(reactant_mol)
        except Exception as e:
            pass

        candidate_smiles = robust_edit_mol(reactant_mol, combo)
        candidate_smiles = set(candidate_smiles)
        candidate_smiles_sanitized = set(
            sanitize_smiles_molvs(smiles) for smiles in candidate_smiles
        )
        if product_smiles_sanitized <= candidate_smiles_sanitized:
            sanitized_correct_rank = min(sanitized_correct_rank, current_rank + 1)
        if product_smiles <= candidate_smiles:
            correct_rank = min(correct_rank, current_rank + 1)

        # If we failed to come up with a new candidate, don't increment the counter!
        if len(set(candidate_smiles_list)) > prev_len_candidate_smiles:
            current_rank += 1

        if correct_rank < max_topk + 1 and sanitized_correct_rank < max_topk + 1:
            break

    for k in topks:
        if correct_rank <= k:
            found_info["top_{:d}".format(k)] = True
        if sanitized_correct_rank <= k:
            found_info["top_{:d}_sanitized".format(k)] = True

    return found_info



from torch_cluster import knn_graph
from torch_geometric.data import HeteroData
from rich import print
from Bio.Data.IUPACData import protein_letters_3to1
from esm import FastaBatchedDataset, pretrained
from collections import defaultdict
from tqdm import tqdm
import torch
import os, sys, pickle, hashlib


# Read PDB and CIF files
def read_structure_file(protein_parser, raw_path, sample_id):
    structure = protein_parser.get_structure(sample_id, raw_path)
    all_res, all_pos, all_atom = [], [], []
    for chains in structure:
        # retrieve all atoms in desired chains
        for chain in chains:
            res = chain.get_residues()
            for res in chain:
                # ignore HETATM records
                if res.id[0] != " ":
                    continue
                for atom in res:
                    all_res.append(res.get_resname())
                    all_pos.append(torch.from_numpy(atom.get_coord()))
                    all_atom.append((atom.get_name(), atom.element))
    all_pos = torch.stack(all_pos, dim=0)
    return all_res, all_atom, all_pos


# Resolution
def filter_resolution(all_res, all_atom, all_pos, protein_resolution):
    protein_resolution = protein_resolution.lower()
    if protein_resolution == "atom":
        atoms_to_keep = None
    elif protein_resolution == "backbone":
        atoms_to_keep = {"N", "CA", "O", "CB"}
    elif protein_resolution == "residue":
        atoms_to_keep = {"CA"}
    else:
        raise Exception(
            f"Invalid resolution {protein_resolution}, expected: 'atom', 'backbone' or 'reside'"
        )

    # this converts atoms to residues or to backbone atoms
    if atoms_to_keep is not None:
        to_keep_idx, atom_names = [], []
        for i, a in enumerate(all_atom):
            if a[0] in atoms_to_keep:  # atom name
                to_keep_idx.append(i)
                atom_names.append(a[1])  # element symbol
        seq = [all_res[i] for i in to_keep_idx]
        pos = all_pos[torch.tensor(to_keep_idx)]
    else:
        atom_names = all_atom
        seq = all_res
        pos = all_pos

    assert pos.shape[0] == len(seq) == len(atom_names)

    return atom_names, seq, pos


# - Download AF2 from uniprots
def download_af_structures():
    pass


# KNN Neighbors / Radius
def compute_graph_edges(data, **kwargs):
    edge_index = knn_graph(data["receptor"].pos, kwargs["knn_size"])
    # data["protein", "contact", "protein"].edge_index = edge_index
    data["receptor", "contact", "receptor"].edge_index = edge_index

    return data


def build_graph(atom_names, seq, pos, sample_id):
    data = HeteroData()
    data["sample_id"] = sample_id
    # retrieve position and compute kNN
    data["receptor"].pos = pos.float()
    data["receptor"].seq = seq  # _seq is residue id
    data["receptor"].atom_names = atom_names  # _atom is atom name

    return data


def get_cache_id(model, labels, sequences):
    input_str = f"{model}-{labels}-{sequences}"
    return hashlib.md5(input_str.encode()).hexdigest()


def precompute_node_embeddings(metadata_json, cache_dir, model_location=None):
    # TODO: Add multi-gpu support
    model, alphabet = pretrained.load_model_and_alphabet(model_location)
    model.eval()
    if torch.cuda.is_available():
        model = model.cuda()
    if not os.path.exists(cache_dir):
        os.makedirs(cache_dir)

    # settings used
    toks_per_batch = 4096
    repr_layers = [33]
    include = "per_tok"
    truncation_seq_length = 1022

    seqs = [d["structure_sequence"] for d in metadata_json]
    if "sample_id" in metadata_json[0]:
        names = [d["sample_id"] for d in metadata_json]
    elif "pdb_id" in metadata_json[0]:
        names = [d["pdb_id"] for d in metadata_json]
    elif "uniprot_id" in metadata_json[0]:
        names = [d["uniprot_id"] for d in metadata_json]

    # already computed all ESM embeddings
    all_embeddings_cache_id = get_cache_id(model, names, seqs)
    all_embeddings_cache_id_file = os.path.join(
        cache_dir, f"{all_embeddings_cache_id}.pt"
    )
    if os.path.exists(all_embeddings_cache_id_file):
        return torch.load(all_embeddings_cache_id_file)

    # check which embeddings are missing (in case of crashes)
    embeddings = defaultdict(dict)
    missing_seqs = []
    missing_names = []
    for i, seq in tqdm(
        enumerate(seqs), desc="Checking for cached embeddings...", total=len(seqs)
    ):
        # This was dumb, can just use the name (uniprot_id)
        # seq_cache_id = get_cache_id(model, [names[i]], [seq])
        seq_cache_file = os.path.join(cache_dir, f"{names[i]}.pt")

        if os.path.exists(seq_cache_file):
            # with open(seq_cache_file, "rb") as f:
            #     seq_embedding = torch.load(f)
            # embeddings[names[i]]["embedding"] = seq_embedding
            embeddings[names[i]] = seq_cache_file
        else:
            missing_seqs.append(seq)
            missing_names.append(names[i])

    dataset = FastaBatchedDataset(missing_names, missing_seqs)
    batches = dataset.get_batch_indices(toks_per_batch, extra_toks_per_seq=1)
    data_loader = torch.utils.data.DataLoader(
        dataset,
        collate_fn=alphabet.get_batch_converter(truncation_seq_length),
        batch_sampler=batches,
    )

    assert all(-(model.num_layers + 1) <= i <= model.num_layers for i in repr_layers)
    repr_layers = [
        (i + model.num_layers + 1) % (model.num_layers + 1) for i in repr_layers
    ]

    with torch.no_grad():
        for batch_idx, (batch_labels, batch_strs, batch_toks) in enumerate(data_loader):
            print(
                f"Processing {batch_idx + 1} of {len(batches)} batches ({batch_toks.size(0)} sequences)"
            )
            if torch.cuda.is_available():
                batch_toks = batch_toks.to(device="cuda", non_blocking=True)

            out = model(batch_toks, repr_layers=repr_layers, return_contacts=False)
            representations = {
                layer: t.to(device="cpu") for layer, t in out["representations"].items()
            }

            for j, sample_id in enumerate(batch_labels):
                truncate_len = min(truncation_seq_length, len(batch_strs[j]))
                seq_embedding = representations[33][j, 1 : truncate_len + 1].clone()
                # embeddings[sample_id]["embedding"] = seq_embedding

                # Cache the embedding for this sequence
                # This was dumb, can just use the name (uniprot_id)
                # seq_cache_id = get_cache_id(model, [sample_id], [batch_strs[j]])
                seq_cache_file = os.path.join(cache_dir, f"{sample_id}.pt")
                embeddings[sample_id] = seq_cache_file

                with open(seq_cache_file, "wb") as f:
                    torch.save(seq_embedding, f)

    # Cache the embeddings paths for all sequences
    with open(all_embeddings_cache_id_file, "wb") as f:
        torch.save(embeddings, f)

    return embeddings


# Node embeddings (needs to be robust, overwritable)
def compute_node_embedding(data, **kwargs):
    # TODO: Add multi-gpu support
    # TODO: Ready to go atomic embeddings
    # Ready to go ESM embeddings
    model = kwargs["model"]
    model_location = kwargs["model_location"]
    alphabet = kwargs["alphabet"]
    batch_converter = kwargs["batch_converter"]

    model.eval()
    # settings used
    toks_per_batch = 4096
    repr_layers = [33]
    truncation_seq_length = 1022

    seq = data.structure_sequence
    batch_labels, batch_strs, batch_tokens = batch_converter([(0, seq)])
    batch_toks = batch_tokens

    assert all(-(model.num_layers + 1) <= i <= model.num_layers for i in repr_layers)

    repr_layers = [
        (i + model.num_layers + 1) % (model.num_layers + 1) for i in repr_layers
    ]

    with torch.no_grad():
        out = model(batch_toks, repr_layers=repr_layers, return_contacts=False)
        representations = {layer: t for layer, t in out["representations"].items()}
        truncate_len = min(truncation_seq_length, len(seq))
        seq_embedding = representations[33][0, 1 : truncate_len + 1].clone()

    return seq_embedding


def get_sequences(
    protein_parser, sample_ids, protein_filepaths, protein_sequences=None
):
    new_sequences = []
    for idx, path in tqdm(
        enumerate(protein_filepaths),
        desc="Reading protein sequences",
        total=len(protein_filepaths),
    ):
        structure_seq = get_sequences_from_structure(protein_parser, path)
        if (
            protein_sequences is not None
            and protein_sequences[idx] is not None
            and structure_seq != protein_sequences[idx]
        ):
            warnings.warn(
                f"Found mismatch in structure seq and metadata seq for sample {sample_ids[idx]['sample_id']}"
            )
        new_sequences.append(structure_seq)
    return new_sequences


def get_sequences_from_structure(protein_parser, file_path):
    structure = protein_parser.get_structure("does_not_matter", file_path)
    structure = structure[0]
    sequence = None
    protein_letters_3to1.update({k.upper(): v for k, v in protein_letters_3to1.items()})
    for i, chain in enumerate(structure):
        seq = ""
        for res_idx, residue in enumerate(chain):
            if residue.get_resname() == "HOH":
                continue
            residue_coords = []
            c_alpha, n, c = None, None, None
            for atom in residue:
                if atom.name == "CA":
                    c_alpha = list(atom.get_vector())
                if atom.name == "N":
                    n = list(atom.get_vector())
                if atom.name == "C":
                    c = list(atom.get_vector())
            if (
                c_alpha != None and n != None and c != None
            ):  # only append residue if it is an amino acid
                try:
                    seq += protein_letters_3to1[residue.get_resname().upper()]
                except Exception as e:
                    seq += "-"
                    print(
                        "encountered unknown AA: ",
                        residue.get_resname(),
                        " in the complex. Replacing it with a dash - .",
                    )

        if sequence is None:
            sequence = seq
        else:
            sequence += ":" + seq

    return sequence


# Process SDF files and SMILES
# def read_ligand_files():
#     pass
# # add ligand graphs or other protein graphs to existing graph
# def combine_graphs():
#     pass
# - Parallel processing
# #     - Complex generation
# #     - ESM embeddings
# #     - AF downloading
# - Can compute intelligent splits
#     - mmseqs split
#     - random split
#     - easy to add additional splitting



import os
from clipzyme.utils.registry import register_object
from clipzyme.utils.classes import Nox
from pytorch_lightning.callbacks import ModelCheckpoint, LearningRateMonitor

# TODO: add args for various callbacks -- currently hardcoded


@register_object("checkpointer", "callback")
class Checkpoint(ModelCheckpoint, Nox):
    def __init__(self, args) -> None:
        super().__init__(
            monitor=args.monitor,
            dirpath=os.path.join(args.checkpoint_dir, args.experiment_name),
            mode="min" if "loss" in args.monitor else "max",
            filename="{}".format(args.experiment_name) + "{epoch}",
            every_n_epochs=1,
            save_top_k=args.checkpoint_save_top_k,
            save_last=args.checkpoint_save_last,
        )


@register_object("lr_monitor", "callback")
class LRMonitor(LearningRateMonitor, Nox):
    def __init__(self, args) -> None:
        super().__init__(logging_interval="step")



import argparse
from typing import Union, Tuple, Any, List
import os, sys
import pickle
import random
import warnings
import itertools
from collections import defaultdict
import csv
import hashlib

# from tqdm import tqdm

import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F
from torch_cluster import knn_graph
from torch_geometric.data import HeteroData
from rich import print as rprint

from scipy.spatial.transform import Rotation

import Bio

import Bio.PDB

warnings.filterwarnings("ignore", category=Bio.PDB.PDBExceptions.PDBConstructionWarning)
from Bio.Data.IUPACData import protein_letters_3to1


def read_files(data_to_load, args):
    pdb_id2prot_dict = {}
    for sample in data_to_load:
        pdb_path = sample["path"]
        pdb_id = sample["sample_id"]
        item = pdb_id2prot_dict.get(pdb_id)
        # avoid recomputing proteins
        if item is None:
            tup_all_res_all_atom_all_pos = parse_pdb(pdb_id, pdb_path)
            sample["receptor"] = tup_all_res_all_atom_all_pos
            pdb_id2prot_dict[pdb_id] = sample

    # write to cache
    data_cache = os.path.join(
        args.protein_cache_path, f"{args.hash_saved_data}_cached_prot_data.pkl"
    )
    if args.debug:
        data_cache = data_cache.replace(".pkl", "_debug.pkl")

    if not args.no_graph_cache and not os.path.exists(data_cache):
        if not os.path.exists(os.path.dirname(data_cache)):
            try:
                os.makedirs(os.path.dirname(data_cache))
            except OSError as exc:  # Guard against race condition
                if exc.errno != exc.errno.EEXIST:
                    raise
        with open(data_cache, "wb+") as f:
            pickle.dump(pdb_id2prot_dict, f)
    return pdb_id2prot_dict


def parse_pdb(pdb_id, pdb_path, models=None, chains=None):
    """
    Parse PDB file via Biopython
    (can cause issues with multi-processing)

    Args:
        pdb_id (str): PDB ID
        pdb_path (str): path to PDB file
        models (list): list of models to select
        chains (list): list of chains to select

    Returns:
        Tuple: (residue sequence, coordinates, atom sequence)
    """
    all_res, all_atom, all_pos = [], [], []
    # biopython PDB parser
    if pdb_path[-3:] == "pdb":
        parser = Bio.PDB.PDBParser()
    elif pdb_path[-3:] == "cif":
        parser = Bio.PDB.MMCIFParser()
    structure = parser.get_structure(pdb_id, pdb_path)  # name, path
    # (optional) select subset of models
    if models is not None:
        models = [structure[m] for m in models]
    else:
        models = structure
    for model in models:
        # (optional) select subset of chains
        if chains is not None:
            chains = [model[c] for c in chains if c in model]
        else:
            chains = model
        # retrieve all atoms in desired chains
        for chain in chains:
            res = chain.get_residues()
            for res in chain:
                # ignore HETATM records
                if res.id[0] != " ":
                    continue
                for atom in res:
                    all_res.append(res.get_resname())
                    all_pos.append(torch.from_numpy(atom.get_coord()))
                    all_atom.append((atom.get_name(), atom.element))
    all_pos = torch.stack(all_pos, dim=0)
    return all_res, all_atom, all_pos


def process_data(pdb_id2prot_dict, args):
    """
    1. Converts protein to resolution of interest (atomic level, backbone, or residue)
    2. Converts protein to HeteroData graph object
    3. Caches prot_dicts and graphs

    Args:
        pdb_id2prot_dict (dict): dictionary of protein data
        args (argparse.Namespace): command line arguments

    Returns:

    """
    # check if cache exists
    graph_cache = os.path.join(
        args.protein_cache_path,
        f"{args.hash_saved_data}_cached_prot_graph_{args.protein_resolution}.pkl",
    )
    if args.debug:
        graph_cache = graph_cache.replace(".pkl", "_debug.pkl")

    if not args.no_graph_cache and os.path.exists(graph_cache):
        with open(graph_cache, "rb") as f:
            pdb_id2prot_dict = pickle.load(f)
            # print("Loaded processed data from cache")
            return pdb_id2prot_dict

    # select subset of residues that match desired resolution (e.g. residue-level vs. atom-level)
    for item in pdb_id2prot_dict.values():
        subset = convert_pdb(*item["receptor"], args)
        item["receptor_atom"] = subset[0]
        item["receptor_seq"] = subset[1]
        item["receptor_xyz"] = subset[2]
        # if len(subset[2]) != len(item['sequence']):
        # return None

    # convert to HeteroData graph objects
    for item in pdb_id2prot_dict.values():
        item["graph"] = to_graph(item, args)

    if not args.no_graph_cache:
        with open(graph_cache, "wb+") as f:
            pickle.dump(pdb_id2prot_dict, f)

    return pdb_id2prot_dict


def convert_pdb(all_res, all_atom, all_pos, args):
    """
    Unify PDB representation across different dataset formats.
    Given all residues, coordinates, and atoms, select subset
    of all atoms to keep.

    Args:
        all_res (list): list of all residue names
        all_atom (list): list of all atom names
        all_pos (torch.Tensor): tensor of all atom coordinates

    Returns:
        tuple(list, torch.Tensor) seq, pos
    """
    if args.protein_resolution == "atom":
        atoms_to_keep = None
    elif args.protein_resolution == "backbone":
        atoms_to_keep = {"N", "CA", "O", "CB"}
    elif args.protein_resolution == "residue":
        atoms_to_keep = {"CA"}
    else:
        raise Exception(f"invalid resolution {args.protein_resolution}")

    # this converts atoms to residues or to backbone atoms
    if atoms_to_keep is not None:
        to_keep_idx, atom_names = [], []
        for i, a in enumerate(all_atom):
            if a[0] in atoms_to_keep:  # atom name
                to_keep_idx.append(i)
                atom_names.append(a[1])  # element symbol
        seq = [all_res[i] for i in to_keep_idx]
        pos = all_pos[torch.tensor(to_keep_idx)]
    else:
        atom_names = all_atom
        seq = all_res
        pos = all_pos

    assert pos.shape[0] == len(seq) == len(atom_names)

    return atom_names, seq, pos


def to_graph(item, args):
    """
    Convert raw dictionary to PyTorch geometric object

    Args:
        item (dict): dictionary of protein data
        args (argparse.Namespace): command-line arguments

    Returns:
        HeteroData: PyTorch geometric object
    """
    data = HeteroData()
    data["name"] = item["sample_id"]
    # retrieve position and compute kNN
    data["receptor"].pos = item["receptor_xyz"].float()
    data["receptor"].x = item["receptor_seq"]  # _seq is residue id
    # kNN graph
    edge_index = knn_graph(data["receptor"].pos, args.knn_size)
    data["receptor", "contact", "receptor"].edge_index = edge_index
    # center receptor at origin (pdb can place protein in strange coordinates)
    center = data["receptor"].pos.mean(dim=0, keepdim=True)
    data["receptor"].pos = data["receptor"].pos - center
    data.center = center  # save old center just in case
    return data


def process_embed(pdb_id2prot_dict, args):
    """
    1. Tokenize protein sequences
    2. Compute ESM embeddings
    3. Cache embeddings

    Args:
        pdb_id2prot_dict (dict): dictionary of protein data
        args (argparse.Namespace): command line arguments

    Returns:
        pdb_id2prot_dict (dict): dictionary of protein data, now with esm node features
        data_params (dict): stores relevant esm parameters

    Raises:
        NotImplementedError: [if not using ESM tokenization]
    """
    data_params = {}
    if args.protein_dim > 0:
        pdb_id2prot_dict = compute_embeddings(pdb_id2prot_dict, args)
        data_params["num_residues"] = 23  # <cls> <sep> <pad>
        print("finished tokenizing residues with ESM")
    else:
        # # Rachel's code for alternative tokenization
        # # tokenize residues for non-ESM
        # tokenizer = tokenize(pdb_id2prot_dict.values(), "receptor_seq")
        # esm_model = None
        # data_params["num_residues"] = len(tokenizer)
        # data_params["tokenizer"] = tokenizer

        # # protein sequence tokenization
        # # tokenize atoms
        # atom_tokenizer = tokenize(pdb_id2prot_dict.values(), "receptor_atom")
        # data_params["atom_tokenizer"] = atom_tokenizer
        # print("finished tokenizing all inputs")
        raise NotImplementedError("Non-ESM tokenization not implemented")

    return pdb_id2prot_dict, data_params


def compute_embeddings(pdb_id2prot_dict, args):
    """
    Pre-compute ESM2 embeddings.
    """
    esm_cache = os.path.join(
        args.protein_cache_path, f"{args.hash_saved_data}_cached_prot_esm.pkl"
    )
    if args.debug:
        esm_cache = esm_cache.replace(".pkl", "_debug.pkl")
    # check if we already computed embeddings
    if not args.no_graph_cache and os.path.exists(esm_cache):
        with open(esm_cache, "rb") as f:
            pdb_id2prot_graph_node_feat = pickle.load(f)
        _save_esm_rep(pdb_id2prot_dict, pdb_id2prot_graph_node_feat)
        print("Loaded cached ESM embeddings")
        return pdb_id2prot_dict

    # print("Computing ESM embeddings")
    # load pretrained model
    torch.hub.set_dir(args.pretrained_hub_dir)
    esm_model, alphabet = torch.hub.load(
        "facebookresearch/esm:main", "esm2_t33_650M_UR50D"
    )
    esm_model = esm_model.cuda().eval()
    tokenizer = alphabet.get_batch_converter()
    # convert to 3 letter codes
    aa_code = defaultdict(lambda: "<unk>")
    aa_code.update({k.upper(): v for k, v in protein_letters_3to1.items()})
    # fix ordering
    pdb_id2prot_dict_sorted = sorted(pdb_id2prot_dict)
    all_graphs = [pdb_id2prot_dict[pdb]["graph"] for pdb in pdb_id2prot_dict_sorted]
    rec_seqs = [g["receptor"].x for g in all_graphs]  # residue ids
    # this is the prot sequence
    rec_seqs = ["".join(aa_code[s] for s in seq) for seq in rec_seqs]
    # batchify sequences
    rec_batches = _esm_batchify(rec_seqs, tokenizer, args)
    with torch.no_grad():
        pad_idx = alphabet.padding_idx
        rec_reps = _run_esm(rec_batches, pad_idx, esm_model)

    # dump to cache
    pdb_id2prot_graph_node_feat = {}
    for idx, pdb in enumerate(pdb_id2prot_dict_sorted):
        # cat one-hot representation and ESM embedding
        # if batch size is 1, unsqueeze to make it 2D
        if len(rec_reps[idx][0].shape) == 1:
            rec_graph_x = torch.cat(
                [rec_reps[idx][0].unsqueeze(-1), rec_reps[idx][1]], dim=1
            )
        else:
            rec_graph_x = torch.cat([rec_reps[idx][0], rec_reps[idx][1]], dim=1)
        pdb_id2prot_graph_node_feat[pdb] = rec_graph_x

    if not args.no_graph_cache:
        with open(esm_cache, "wb+") as f:
            pickle.dump(pdb_id2prot_graph_node_feat, f)

    # overwrite graph.x for each element in batch
    _save_esm_rep(pdb_id2prot_dict, pdb_id2prot_graph_node_feat)

    return pdb_id2prot_dict


def _save_esm_rep(pdb_id2prot_dict, pdb_id2prot_graph_node_feat):
    """
    Assign new ESM representation to graph.x INPLACE
    """
    for pdb_id, prot_graph_node_features in pdb_id2prot_graph_node_feat.items():
        rec_graph = pdb_id2prot_dict[pdb_id]["graph"]["receptor"]
        rec_graph.x = prot_graph_node_features
        # assert len(rec_graph.pos) == len(rec_graph.x)
        if not len(rec_graph.pos) == len(rec_graph.x):
            print(f"Sample's coords do not match sequence! {pdb_id2prot_dict}")
            pdb_id2prot_dict = None
            break

    return pdb_id2prot_dict


def _esm_batchify(seqs, tokenizer, args):
    batch_size = args.batch_size
    # group up sequences
    batches = [seqs[i : i + batch_size] for i in range(0, len(seqs), batch_size)]
    batches = [[("", seq) for seq in batch] for batch in batches]
    # tokenize
    batch_tokens = [tokenizer(batch)[2] for batch in batches]
    return batch_tokens


def _run_esm(batches, padding_idx, esm_model):
    """
    Wrapper around ESM specifics

    Args:
        batches (list): list of tokenized sequences
        padding_idx (int): padding index
        esm_model (nn.Module): ESM model
    Returns:
        list: list of (esm_repr, tokens) without <CLS> and <EOS>
    """
    # run ESM model
    all_reps = []
    for batch in batches:
        reps = esm_model(batch.cuda(), repr_layers=[33])
        reps = reps["representations"][33].cpu().squeeze()[:, 1:]
        all_reps.append(reps)
    # crop to length
    # exclude <cls> <eos>
    cropped_representations = []
    for i, batch in enumerate(batches):
        batch_lens = (batch != padding_idx).sum(1) - 2
        for j, length in enumerate(batch_lens):
            rep_crop = all_reps[i][j, :length]
            token_crop = batch[j, 1 : length + 1, None]
            cropped_representations.append((rep_crop, token_crop))
    return cropped_representations


def tokenize(data, key, tokenizer=None):
    """
    Tokenize every item[key] in data.
    Modifies item[key] and copies original value to item[key_raw].
    If tokenizer is None, creates a new tokenizer.
    If tokenizer is a dict, uses it to tokenize.

    Args:
        data (list): list of dicts (items)
        key (str): key to tokenize in each item[key]

    Returns:
        dict: tokenizer
    """
    if len(data) == 0:  # sometimes no val split, etc.
        return
    # if tokenizer is not provided, create index
    all_values = [item[key] for item in data]
    all_values = set(itertools.chain(*all_values))
    if tokenizer is None:
        tokenizer = {}  # never used
    if type(tokenizer) is dict:
        for item in sorted(all_values):
            if item not in tokenizer:
                tokenizer[item] = len(tokenizer) + 1  # 1-index
        f_token = lambda seq: [tokenizer[x] for x in seq]
    else:
        aa_code = defaultdict(lambda: "<unk>")
        aa_code.update(
            {three.upper(): one for three, one in protein_letters_3to1.items()}
        )

        def f_token(seq):
            seq = "".join([aa_code[s] for s in seq])
            seq = tokenizer([("", seq)])[2][0]
            return seq

    # tokenize items and modify data in place
    raw_key = f"{key}_raw"
    for item in data:
        raw_item = item[key]
        item[raw_key] = raw_item  # save old
        # tokenize and convert to tensor if applicable
        item[key] = f_token(raw_item)
        if not torch.is_tensor(item[key]):
            item[key] = torch.tensor(item[key])
    return tokenizer


def get_protein_graphs_from_path(data_to_load, args):
    """
     Args:
        data_to_load (list): list of dicts with key 'path'
        args (argparse.ArgumentParser)

    Raises:
        NotImplementedError: [if not using ESM tokenization]
    """
    # cache for post-processed data, optional
    args.hash_saved_data = hashlib.md5(f"{data_to_load}".encode()).hexdigest()

    # load and process files
    # organizes and caches pdb_id2prot_dict
    pdb_id2prot_dict = read_files(data_to_load, args)

    # filters resolution, generates graphs and caches pdb_id2prot_dict
    data_params = {}
    pdb_id2prot_dict = process_data(pdb_id2prot_dict, args)
    if pdb_id2prot_dict is None:
        return None, None
    # generate ESM embeddings and cache cat([one-hot, esm_embeddings])
    # pdb_id2prot_dict, data_params = process_embed(pdb_id2prot_dict, args)
    sample = [pdb_id2prot_dict[key] for key in pdb_id2prot_dict][0]
    return sample, data_params
